import datetime
import os
import pkg_resources
from dask.distributed import Client
from logzero import logger

import covigator
from covigator.accessor.ena_accessor import EnaAccessor
from covigator.configuration import Configuration
from covigator.database.database import Database
from covigator.database.model import DataSource, JobStatus, SampleGisaid, SampleEna
from covigator.database.queries import Queries
from covigator.processor.abstract_processor import AbstractProcessor
from covigator.tests import SARS_COV_2_TAXID, HOMO_SAPIENS_TAXID


class FakeConfiguration(Configuration):

    def __init__(self):
        # use a folder with write permissions for testing
        os.environ[self.ENV_COVIGATOR_STORAGE_FOLDER] = "./data/covigator"
        os.environ[self.ENV_COVIGATOR_REF_FASTA] = pkg_resources.resource_filename(
            covigator.tests.__name__, "resources/MN908947.3.fa")

        # this makes sure that we do not wipe a relevant database by mistake
        os.environ[self.ENV_COVIGATOR_TABLE_VERSION] = "_test"

        super().__init__()


class FakeEnaAccessor(EnaAccessor):

    def __init__(self, results, database=None):
        # uses an in memory database or the one provided
        super().__init__(tax_id=SARS_COV_2_TAXID, host_tax_id=HOMO_SAPIENS_TAXID,
                         database=database if database else Database(test=True, config=Configuration()))
        self.results = results

    def _get_ena_runs_page(self):
        return self.results


class FakeEnaProcessor(AbstractProcessor):

    def __init__(self, database: Database, dask_client: Client, config: Configuration):
        logger.info("Initialising ENA processor")
        super().__init__(database, dask_client, DataSource.ENA, config, wait_time=1)

    def _process_run(self, run_accession: str):
        """
        Launches all jobs and returns the futures for the final job only
        """
        # NOTE: here we set the priority of each step to ensure a depth first processing
        future = self.dask_client.submit(FakeEnaProcessor.job, self.config, run_accession, priority=1)
        return future

    @staticmethod
    def job(config: Configuration, run_accession):
        return FakeEnaProcessor.run_job(
            config, run_accession, start_status=JobStatus.QUEUED, end_status=JobStatus.FINISHED,
            error_status=JobStatus.FAILED_PROCESSING, data_source=DataSource.ENA,
            function=FakeEnaProcessor.run_all
    )

    @staticmethod
    def run_all(sample: SampleEna, queries: Queries, config: Configuration) -> SampleEna:
        logger.info("Job processed!")
        sample.analysed_at = datetime.datetime.now()
        sample.pangolin_lineage = "B.TEST"
        return sample


class FakeProcessorFailing(AbstractProcessor):

    def __init__(self, database: Database, dask_client: Client, config: Configuration, source: DataSource):
        logger.info("Initialising fake failing processor")
        super().__init__(database, dask_client, source, config, wait_time=1)
        self.source = source

    def _process_run(self, run_accession: str):
        """
        Launches all jobs and returns the futures for the final job only
        """
        # NOTE: here we set the priority of each step to ensure a depth first processing
        future = self.dask_client.submit(FakeProcessorFailing.job, self.config, run_accession, self.source, priority=1)
        return future

    @staticmethod
    def job(config: Configuration, run_accession, source: DataSource):
        return FakeProcessorFailing.run_job(
            config, run_accession, start_status=JobStatus.QUEUED, end_status=JobStatus.FINISHED,
            error_status=JobStatus.FAILED_PROCESSING, data_source=source,
            function=FakeProcessorFailing.run_all
        )

    @staticmethod
    def run_all(sample: SampleEna, queries: Queries, config: Configuration) -> SampleEna:
        raise ValueError("Fail em'all")


class FakeGisaidProcessor(AbstractProcessor):

    def __init__(self, database: Database, dask_client: Client, config: Configuration):
        logger.info("Initialising GISAID processor")
        super().__init__(database, dask_client, DataSource.GISAID, config, wait_time=1)

    def _process_run(self, run_accession: str):
        """
        Launches all jobs and returns the futures for the final job only
        """
        # NOTE: here we set the priority of each step to ensure a depth first processing
        future = self.dask_client.submit(FakeGisaidProcessor.job, self.config, run_accession, priority=1)
        return future

    @staticmethod
    def job(config: Configuration, run_accession):
        return FakeGisaidProcessor.run_job(
            config, run_accession, start_status=JobStatus.QUEUED, end_status=JobStatus.FINISHED,
            error_status=JobStatus.FAILED_PROCESSING, data_source=DataSource.GISAID,
            function=FakeGisaidProcessor.run_all
    )

    @staticmethod
    def run_all(sample: SampleGisaid, queries: Queries, config: Configuration) -> SampleGisaid:
        logger.info("Job processed!")
        sample.analysed_at = datetime.datetime.now()
        sample.pangolin_lineage = "B.TEST"
        return sample



import os
from datetime import datetime
from typing import List
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, String, Float, Enum, DateTime, Integer, Boolean, Date, ForeignKey, \
    ForeignKeyConstraint, BigInteger, JSON, Index
import enum

from covigator.configuration import Configuration


def get_table_versioned_name(basename, config: Configuration):
    return "{}{}".format(basename, config.db_table_version)


config = Configuration()
GENE_TABLE_NAME = get_table_versioned_name('gene', config=config)
DOMAIN_TABLE_NAME = get_table_versioned_name('domain', config=config)
LOG_TABLE_NAME = get_table_versioned_name('log', config=config)
LAST_UPDATE_TABLE_NAME = get_table_versioned_name('last_update', config=config)
VARIANT_COOCCURRENCE_TABLE_NAME = get_table_versioned_name('variant_cooccurrence', config=config)
VARIANT_OBSERVATION_TABLE_NAME = get_table_versioned_name('variant_observation', config=config)
SUBCLONAL_VARIANT_OBSERVATION_TABLE_NAME = get_table_versioned_name('subclonal_variant_observation', config=config)
LOW_FREQUENCY_VARIANT_OBSERVATION_TABLE_NAME = get_table_versioned_name('low_frequency_variant_observation', config=config)
GISAID_VARIANT_OBSERVATION_TABLE_NAME = get_table_versioned_name('gisaid_variant_observation', config=config)
VARIANT_TABLE_NAME = get_table_versioned_name('variant', config=config)
SUBCLONAL_VARIANT_TABLE_NAME = get_table_versioned_name('subclonal_variant', config=config)
LOW_FREQUENCY_VARIANT_TABLE_NAME = get_table_versioned_name('low_frequency_variant', config=config)
GISAID_VARIANT_TABLE_NAME = get_table_versioned_name('gisaid_variant', config=config)
SAMPLE_GISAID_TABLE_NAME = get_table_versioned_name('sample_gisaid', config=config)
SAMPLE_ENA_TABLE_NAME = get_table_versioned_name('sample_ena', config=config)
CONSERVATION_TABLE_NAME = get_table_versioned_name('conservation', config=config)
PRECOMPUTED_VARIANTS_PER_SAMPLE_TABLE_NAME = get_table_versioned_name('precomputed_variants_per_sample', config=config)
PRECOMPUTED_SUBSTITUTIONS_COUNTS_TABLE_NAME = get_table_versioned_name('precomputed_substitutions_counts', config=config)
PRECOMPUTED_INDEL_LENGTH_TABLE_NAME = get_table_versioned_name('precomputed_indel_length', config=config)
PRECOMPUTED_ANNOTATION_TABLE_NAME = get_table_versioned_name('precomputed_annotation', config=config)
PRECOMPUTED_OCCURRENCE_TABLE_NAME = get_table_versioned_name('precomputed_top_occurrence', config=config)
PRECOMPUTED_NS_S_COUNTS_TABLE_NAME = get_table_versioned_name('precomputed_ns_s_counts', config=config)
PRECOMPUTED_DN_DS_BY_DOMAIN_TABLE_NAME = get_table_versioned_name('precomputed_dn_ds_by_domain', config=config)
PRECOMPUTED_TABLE_COUNTS_TABLE_NAME = get_table_versioned_name('precomputed_table_counts', config=config)
PRECOMPUTED_VARIANT_ABUNDANCE_HIST_TABLE_NAME = get_table_versioned_name('precomputed_variant_abundance_histogram', config=config)
PRECOMPUTED_VARIANTS_PER_LINEAGE_TABLE_NAME = get_table_versioned_name('precomputed_variants_per_lineage', config=config)
JOB_STATUS_CONSTRAINT_NAME = get_table_versioned_name('job_status', config=config)
DATA_SOURCE_CONSTRAINT_NAME = get_table_versioned_name('data_source', config=config)
COVIGATOR_MODULE_CONSTRAINT_NAME = get_table_versioned_name('covigator_module', config=config)
REGION_TYPE_CONSTRAINT_NAME = get_table_versioned_name('region_type', config=config)
VARIANT_TYPE_CONSTRAINT_NAME = get_table_versioned_name('variant_type', config=config)
SEPARATOR = ";"

Base = declarative_base()


class Gene(Base):
    """
    This table holds the genes in the organism genome and its annotations in the field `data`. The annotations
    are in JSON format. This data is fetched from ftp://ftp.ensemblgenomes.org/pub/viruses/json/sars_cov_2/sars_cov_2.json
    """
    __tablename__ = GENE_TABLE_NAME

    identifier = Column(String, primary_key=True)
    name = Column(String)
    start = Column(Integer, index=True)
    end = Column(Integer)
    fraction_synonymous = Column(Float)
    fraction_non_synonymous = Column(Float)


class Domain(Base):
    """
    This table holds the Pfam domains for the genes
    """
    __tablename__ = DOMAIN_TABLE_NAME

    name = Column(String, primary_key=True)
    description = Column(String)
    start = Column(Integer, index=True)
    end = Column(Integer)
    fraction_synonymous = Column(Float)
    fraction_non_synonymous = Column(Float)
    gene_identifier = Column(ForeignKey("{}.identifier".format(Gene.__tablename__)))
    gene_name = Column(String)


class JobStatus(enum.Enum):
    """
    Valid job status
    """
    __constraint_name__ = JOB_STATUS_CONSTRAINT_NAME

    PENDING = 1
    FAILED_DOWNLOAD = 2
    DOWNLOADED = 3
    QUEUED = 4
    FAILED_PROCESSING = 5
    FINISHED = 6
    HOLD = 7
    EXCLUDED = 8


class DataSource(enum.Enum):
    """
    Valid sources of data
    """
    __constraint_name__ = DATA_SOURCE_CONSTRAINT_NAME

    ENA = 1
    GISAID = 2


class VariantType(enum.Enum):
    __constraint_name__ = VARIANT_TYPE_CONSTRAINT_NAME

    SNV = 1
    INSERTION = 2
    DELETION = 3
    MNV = 4


class SampleGisaid(Base):
    """
    The table that holds all metadata for a GISAID sample
    """
    __tablename__ = SAMPLE_GISAID_TABLE_NAME

    run_accession = Column(String, primary_key=True)
    # DEPRECATED
    finished = Column(Boolean)
    collection_date = Column(Date)
    # Host information
    host_tax_id = Column(String)
    host = Column(String)
    # geographical data
    country_raw = Column(String)
    region = Column(String)
    country = Column(String)
    country_alpha_2 = Column(String)
    country_alpha_3 = Column(String)
    continent = Column(String)
    continent_alpha_2 = Column(String)
    site = Column(String)
    site2 = Column(String)

    # sequence  information
    sequence = Column(JSON)
    sequence_length = Column(Integer)
    count_n_bases = Column(Integer)
    count_ambiguous_bases = Column(Integer)

    # counts of variants
    count_snvs = Column(Integer)
    count_insertions = Column(Integer)
    count_deletions = Column(Integer)

    # job status
    status = Column(Enum(JobStatus, name=JobStatus.__constraint_name__), default=JobStatus.PENDING)
    created_at = Column(DateTime(timezone=True), nullable=False, default=datetime.now())
    queued_at = Column(DateTime(timezone=True))
    analysed_at = Column(DateTime(timezone=True))
    loaded_at = Column(DateTime(timezone=True))
    failed_at = Column(DateTime(timezone=True))
    error_message = Column(String)

    # output files
    sample_folder = Column(String)
    vcf_path = Column(String)
    fasta_path = Column(String)
    pangolin_path = Column(String)

    # pango output (corresponding only to LoFreq)
    pangolin_lineage = Column(String)
    pangolin_conflict = Column(Float)
    pangolin_ambiguity_score = Column(Float)
    pangolin_scorpio_call = Column(String)
    pangolin_scorpio_support = Column(Float)
    pangolin_scorpio_conflict = Column(Float)
    pangolin_version = Column(String)
    pangolin_pangolin_version = Column(String)
    # TODO: remove upper case from here, it confuses SQL for some operations
    pangolin_pangoLEARN_version = Column(String)
    pangolin_pango_version = Column(String)
    pangolin_status = Column(String)
    pangolin_note = Column(String)

    covigator_accessor_version = Column(String)
    covigator_processor_version = Column(String)

    def get_sample_folder(self, base_folder):
        return os.path.join(
            base_folder,
            self.collection_date.strftime("%Y%m%d") if self.collection_date is not None else "nodate",
            self.run_accession)


class SampleEna(Base):
    """
    The table that holds all metadata for a ENA sample
    """
    __tablename__ = SAMPLE_ENA_TABLE_NAME

    # data on run
    # TODO: add foreign keys to jobs
    run_accession = Column(String, primary_key=True)            # 'ERR4080473',
    # DEPRECATED
    finished = Column(Boolean)
    sample_accession = Column(String)                           # 'SAMEA6798401',
    scientific_name = Column(String)
    study_accession = Column(String)
    experiment_accession = Column(String)
    # TODO: store these as dates
    first_created = Column(Date)
    collection_date = Column(Date)
    instrument_platform = Column(String)
    instrument_model = Column(String)
    library_name = Column(String)
    nominal_length = Column(Integer)
    library_layout = Column(String)
    library_strategy = Column(String)
    library_source = Column(String)
    library_selection = Column(String)
    read_count = Column(BigInteger)
    base_count = Column(BigInteger)
    sample_collection = Column(String)
    sequencing_method = Column(String)
    center_name = Column(String)

    # FASTQs
    fastq_ftp = Column(String, nullable=False)                  # 'ftp.sra.ebi.ac.uk/vol1/fastq/ERR408/003/ERR4080473/ERR4080473_1.fastq.gz',
    fastq_md5 = Column(String, nullable=False)
    num_fastqs = Column(Integer, nullable=False)

    # data on host
    host_tax_id = Column(String)                                # '9606',
    host_sex = Column(String)
    host_body_site = Column(String)
    host_gravidity = Column(String)
    host_phenotype = Column(String)
    host_genotype = Column(String)

    # geographical data
    lat = Column(Float)
    lon = Column(Float)
    country_raw = Column(String)                                    # 'Denmark',
    country = Column(String)
    country_alpha_2 = Column(String)
    country_alpha_3 = Column(String)
    continent = Column(String)
    continent_alpha_2 = Column(String)

    # counts of variants
    count_snvs = Column(Integer)
    count_insertions = Column(Integer)
    count_deletions = Column(Integer)
    count_subclonal_snvs = Column(Integer)
    count_subclonal_insertions = Column(Integer)
    count_subclonal_deletions = Column(Integer)
    count_low_frequency_snvs = Column(Integer)
    count_low_frequency_insertions = Column(Integer)
    count_low_frequency_deletions = Column(Integer)

    # job status
    status = Column(Enum(JobStatus, name=JobStatus.__constraint_name__), default=JobStatus.PENDING)
    created_at = Column(DateTime(timezone=True), nullable=False, default=datetime.now())
    queued_at = Column(DateTime(timezone=True))
    downloaded_at = Column(DateTime(timezone=True))
    analysed_at = Column(DateTime(timezone=True))
    cleaned_at = Column(DateTime(timezone=True))
    loaded_at = Column(DateTime(timezone=True))
    cooccurrence_at = Column(DateTime(timezone=True))
    failed_at = Column(DateTime(timezone=True))
    error_message = Column(String)

    # local files storage
    sample_folder = Column(String)
    fastq_path = Column(String)  # the local path where FASTQ files are stored in semi colon separated list
    lofreq_vcf_path = Column(String)
    ivar_vcf_path = Column(String)
    gatk_vcf_path = Column(String)
    bcftools_vcf_path = Column(String)
    lofreq_pangolin_path = Column(String)
    ivar_pangolin_path = Column(String)
    gatk_pangolin_path = Column(String)
    bcftools_pangolin_path = Column(String)
    fastp_path = Column(String)
    deduplication_metrics_path = Column(String)
    horizontal_coverage_path = Column(String)
    vertical_coverage_path = Column(String)

    # FASTP results
    qc = Column(JSON)

    # horizontal coverage analysis results
    num_reads = Column(Integer)
    covered_bases = Column(Integer)
    coverage = Column(Float)
    mean_depth = Column(Float)
    mean_base_quality = Column(Float)
    mean_mapping_quality = Column(Float)

    # pango output (corresponding only to LoFreq)
    pangolin_lineage = Column(String)
    pangolin_conflict = Column(Float)
    pangolin_ambiguity_score = Column(Float)
    pangolin_scorpio_call = Column(String)
    pangolin_scorpio_support = Column(Float)
    pangolin_scorpio_conflict = Column(Float)
    pangolin_version = Column(String)
    pangolin_pangolin_version = Column(String)
    # TODO: remove upper case from here, it confuses SQL for some operations
    pangolin_pangoLEARN_version = Column(String)
    pangolin_pango_version = Column(String)
    pangolin_status = Column(String)
    pangolin_note = Column(String)

    # Picard deduplicatio output
    percent_duplication = Column(Float)
    unpaired_reads_examined = Column(Integer)
    read_pairs_examined = Column(Integer)
    secondary_or_supplementary_reads = Column(Integer)
    unmapped_reads = Column(Integer)
    unpaired_read_duplicates = Column(Integer)
    read_pair_duplicates = Column(Integer)
    read_pair_optical_duplicates = Column(Integer)

    covigator_accessor_version = Column(String)
    covigator_processor_version = Column(String)

    def get_sample_folder(self, base_folder):
        return os.path.join(
            base_folder,
            self.collection_date.strftime("%Y%m%d") if self.collection_date is not None else "nodate",
            self.run_accession)

    def get_fastqs_ftp(self) -> List:
        return self.fastq_ftp.split(SEPARATOR) if self.fastq_ftp is not None else []

    def get_fastqs_md5(self) -> List:
        return self.fastq_md5.split(SEPARATOR) if self.fastq_md5 is not None else []

    def get_fastq_paths(self) -> List:
        return self.fastq_path.split(SEPARATOR) if self.fastq_path is not None else []

    def get_fastq1_and_fastq2(self):
        fastqs = self.get_fastq_paths()
        if len(fastqs) > 1:
            fastq1 = next(filter(lambda x: "_1.fastq" in x, fastqs))
            fastq2 = next(filter(lambda x: "_2.fastq" in x, fastqs))
        else:
            fastq1 = fastqs[0]
            fastq2 = None
        return fastq1, fastq2


class Variant(Base):
    """
    A variant with its specific annotations. THis does not contain any sample specific annotations.
    """
    __tablename__ = VARIANT_TABLE_NAME

    variant_id = Column(String, primary_key=True)
    chromosome = Column(String)
    position = Column(Integer, index=True)
    reference = Column(String)
    alternate = Column(String)
    overlaps_multiple_genes = Column(Boolean, default=False)
    """
    SnpEff annotations
    ##INFO=<ID=ANN,Number=.,Type=String,Description="Functional annotations: '
    Allele |                    C 
    Annotation |                missense_variant
    Annotation_Impact |         MODERATE 
    Gene_Name |                 orf1ab 
    Gene_ID |                   gene-GU280_gp01 
    Feature_Type |              transcript 
    Feature_ID |                TRANSCRIPT_gene-GU280_gp01 
    Transcript_BioType |        protein_coding 
    Rank |                      1/1 
    HGVS.c |                    c.33A>C 
    HGVS.p |                    p.Lys11Asn 
    cDNA.pos / cDNA.length |    33/21290 
    CDS.pos / CDS.length |      33/21290 
    AA.pos / AA.length |        11/7095 
    Distance | 
    ERRORS / WARNINGS / INFO'   
    """
    annotation = Column(String, index=True)
    annotation_highest_impact = Column(String, index=True)
    annotation_impact = Column(String, index=True)
    gene_name = Column(String, index=True)
    gene_id = Column(String)
    biotype = Column(String)
    hgvs_c = Column(String)
    hgvs_p = Column(String)
    cdna_pos_length = Column(String)
    cds_pos_length = Column(String)
    aa_pos_length = Column(String)

    # derived annotations
    variant_type = Column(Enum(VariantType, name=VariantType.__constraint_name__))
    length = Column(Integer)
    reference_amino_acid = Column(String)
    alternate_amino_acid = Column(String)
    position_amino_acid = Column(Integer)

    # ConsHMM conservation annotations
    cons_hmm_sars_cov_2 = Column(Float)
    cons_hmm_sarbecovirus = Column(Float)
    cons_hmm_vertebrate_cov = Column(Float)

    # Pfam protein domains
    pfam_name = Column(String)
    pfam_description = Column(String)

    def get_variant_id(self):
        return "{}:{}>{}".format(self.position, self.reference, self.alternate)


class SubclonalVariant(Base):

    __tablename__ = SUBCLONAL_VARIANT_TABLE_NAME

    variant_id = Column(String, primary_key=True)
    chromosome = Column(String)
    position = Column(Integer, index=True)
    reference = Column(String)
    alternate = Column(String)
    overlaps_multiple_genes = Column(Boolean, default=False)

    annotation = Column(String, index=True)
    annotation_highest_impact = Column(String, index=True)
    annotation_impact = Column(String, index=True)
    gene_name = Column(String, index=True)
    gene_id = Column(String)
    biotype = Column(String)
    hgvs_c = Column(String)
    hgvs_p = Column(String)
    cdna_pos_length = Column(String)
    cds_pos_length = Column(String)
    aa_pos_length = Column(String)

    # derived annotations
    variant_type = Column(Enum(VariantType, name=VariantType.__constraint_name__))
    length = Column(Integer)
    reference_amino_acid = Column(String)
    alternate_amino_acid = Column(String)
    position_amino_acid = Column(Integer)

    # ConsHMM conservation annotations
    cons_hmm_sars_cov_2 = Column(Float)
    cons_hmm_sarbecovirus = Column(Float)
    cons_hmm_vertebrate_cov = Column(Float)

    # Pfam protein domains
    pfam_name = Column(String)
    pfam_description = Column(String)

    def get_variant_id(self):
        return "{}:{}>{}".format(self.position, self.reference, self.alternate)


class LowFrequencyVariant(Base):
    __tablename__ = LOW_FREQUENCY_VARIANT_TABLE_NAME

    variant_id = Column(String, primary_key=True)
    chromosome = Column(String)
    position = Column(Integer, index=True)
    reference = Column(String)
    alternate = Column(String)
    overlaps_multiple_genes = Column(Boolean, default=False)

    annotation = Column(String, index=True)
    annotation_highest_impact = Column(String, index=True)
    annotation_impact = Column(String, index=True)
    gene_name = Column(String, index=True)
    gene_id = Column(String)
    biotype = Column(String)
    hgvs_c = Column(String)
    hgvs_p = Column(String)
    cdna_pos_length = Column(String)
    cds_pos_length = Column(String)
    aa_pos_length = Column(String)

    # derived annotations
    variant_type = Column(Enum(VariantType, name=VariantType.__constraint_name__))
    length = Column(Integer)
    reference_amino_acid = Column(String)
    alternate_amino_acid = Column(String)
    position_amino_acid = Column(Integer)

    # ConsHMM conservation annotations
    cons_hmm_sars_cov_2 = Column(Float)
    cons_hmm_sarbecovirus = Column(Float)
    cons_hmm_vertebrate_cov = Column(Float)

    # Pfam protein domains
    pfam_name = Column(String)
    pfam_description = Column(String)

    def get_variant_id(self):
        return "{}:{}>{}".format(self.position, self.reference, self.alternate)


class GisaidVariant(Base):
    __tablename__ = GISAID_VARIANT_TABLE_NAME

    variant_id = Column(String, primary_key=True)
    chromosome = Column(String)
    position = Column(Integer, index=True)
    reference = Column(String)
    alternate = Column(String)
    overlaps_multiple_genes = Column(Boolean, default=False)

    annotation = Column(String, index=True)
    annotation_highest_impact = Column(String, index=True)
    annotation_impact = Column(String, index=True)
    gene_name = Column(String, index=True)
    gene_id = Column(String)
    biotype = Column(String)
    hgvs_c = Column(String)
    hgvs_p = Column(String)
    cdna_pos_length = Column(String)
    cds_pos_length = Column(String)
    aa_pos_length = Column(String)

    # derived annotations
    variant_type = Column(Enum(VariantType, name=VariantType.__constraint_name__))
    length = Column(Integer)
    reference_amino_acid = Column(String)
    alternate_amino_acid = Column(String)
    position_amino_acid = Column(Integer)

    # ConsHMM conservation annotations
    cons_hmm_sars_cov_2 = Column(Float)
    cons_hmm_sarbecovirus = Column(Float)
    cons_hmm_vertebrate_cov = Column(Float)

    # Pfam protein domains
    pfam_name = Column(String)
    pfam_description = Column(String)

    def get_variant_id(self):
        return "{}:{}>{}".format(self.position, self.reference, self.alternate)


class VariantObservation(Base):
    """
    A variant observation in a particular sample. This contains all annotations of a specific observation of a variant.
    """
    __tablename__ = VARIANT_OBSERVATION_TABLE_NAME

    sample = Column(String, primary_key=True)
    variant_id = Column(String, primary_key=True)
    chromosome = Column(String)
    position = Column(Integer)
    reference = Column(String)
    alternate = Column(String)
    quality = Column(Float)
    filter = Column(String)
    """
    ##INFO=<ID=DP,Number=1,Type=Integer,Description="Raw Depth">
    ##INFO=<ID=AF,Number=1,Type=Float,Description="Allele Frequency">
    ##INFO=<ID=SB,Number=1,Type=Integer,Description="Phred-scaled strand bias at this position">
    ##INFO=<ID=DP4,Number=4,Type=Integer,Description="Counts for ref-forward bases, ref-reverse, alt-forward and alt-reverse bases">
    ##INFO=<ID=INDEL,Number=0,Type=Flag,Description="Indicates that the variant is an INDEL.">
    ##INFO=<ID=CONSVAR,Number=0,Type=Flag,Description="Indicates that the variant is a consensus variant (as opposed to a low frequency variant).">
    ##INFO=<ID=HRUN,Number=1,Type=Integer,Description="Homopolymer length to the right of report indel position">
    """
    dp = Column(Integer)
    ac = Column(Integer)
    dp4_ref_forward = Column(Integer)
    dp4_ref_reverse = Column(Integer)
    dp4_alt_forward = Column(Integer)
    dp4_alt_reverse = Column(Integer)
    vaf = Column(Float)
    strand_bias = Column(Integer)
    # fields replicated from Variant for performance reasons
    annotation = Column(String)
    annotation_impact = Column(String)
    biotype = Column(String)
    annotation_highest_impact = Column(String)
    gene_name = Column(String)
    hgvs_c = Column(String)
    hgvs_p = Column(String)

    # fields replicated from sample for performance reasons
    date = Column(Date)

    # derived annotations
    variant_type = Column(Enum(VariantType, name=VariantType.__constraint_name__))
    length = Column(Integer)
    reference_amino_acid = Column(String)
    alternate_amino_acid = Column(String)
    position_amino_acid = Column(Integer)

    # ConsHMM conservation annotations
    cons_hmm_sars_cov_2 = Column(Float)
    cons_hmm_sarbecovirus = Column(Float)
    cons_hmm_vertebrate_cov = Column(Float)

    # Pfam protein domains
    pfam_name = Column(String)
    pfam_description = Column(String)

    ForeignKeyConstraint([sample], [SampleEna.run_accession])
    ForeignKeyConstraint([variant_id], [Variant.variant_id])

    __table_args__ = (
        Index("{}_index_annotation_position".format(VARIANT_OBSERVATION_TABLE_NAME),
              "annotation_highest_impact", "position"),
        Index("{}_index_sample".format(VARIANT_OBSERVATION_TABLE_NAME), "sample"),
        Index("{}_index_position".format(VARIANT_OBSERVATION_TABLE_NAME), "position"),
        Index("{}_index_annotation".format(VARIANT_OBSERVATION_TABLE_NAME), "annotation_highest_impact"),
        Index("{}_index_variant_id".format(VARIANT_OBSERVATION_TABLE_NAME), "variant_id")
    )


class SubclonalVariantObservation(Base):
    """
    A variant observation in a particular sample. This contains all annotations of a specific observation of a variant.
    """
    __tablename__ = SUBCLONAL_VARIANT_OBSERVATION_TABLE_NAME

    sample = Column(String, primary_key=True)
    variant_id = Column(String, primary_key=True)
    chromosome = Column(String)
    position = Column(Integer)
    reference = Column(String)
    alternate = Column(String)
    quality = Column(Float)
    filter = Column(String)
    dp = Column(Integer)
    ac = Column(Integer)
    dp4_ref_forward = Column(Integer)
    dp4_ref_reverse = Column(Integer)
    dp4_alt_forward = Column(Integer)
    dp4_alt_reverse = Column(Integer)
    vaf = Column(Float)
    strand_bias = Column(Integer)

    # fields replicated from Variant for performance reasons
    annotation = Column(String)
    annotation_impact = Column(String)
    biotype = Column(String)
    annotation_highest_impact = Column(String, index=True)
    gene_name = Column(String)
    hgvs_c = Column(String)
    hgvs_p = Column(String)

    # fields replicated from sample for performance reasons
    date = Column(Date)

    # derived annotations
    variant_type = Column(Enum(VariantType, name=VariantType.__constraint_name__))
    length = Column(Integer)
    reference_amino_acid = Column(String)
    alternate_amino_acid = Column(String)
    position_amino_acid = Column(Integer)

    # ConsHMM conservation annotations
    cons_hmm_sars_cov_2 = Column(Float)
    cons_hmm_sarbecovirus = Column(Float)
    cons_hmm_vertebrate_cov = Column(Float)

    # Pfam protein domains
    pfam_name = Column(String)
    pfam_description = Column(String)

    ForeignKeyConstraint([sample], [SampleEna.run_accession])
    ForeignKeyConstraint([variant_id], [SubclonalVariant.variant_id])

    __table_args__ = (
        Index("{}_index_position".format(SUBCLONAL_VARIANT_OBSERVATION_TABLE_NAME), "position"),
        Index("{}_index_annotation_vaf".format(SUBCLONAL_VARIANT_OBSERVATION_TABLE_NAME), "annotation_highest_impact", "vaf"),
        Index("{}_index_vaf".format(SUBCLONAL_VARIANT_OBSERVATION_TABLE_NAME), "vaf"),
    )


class LowFrequencyVariantObservation(Base):
    """
    A variant observation in a particular sample. This contains all annotations of a specific observation of a variant.
    """
    __tablename__ = LOW_FREQUENCY_VARIANT_OBSERVATION_TABLE_NAME

    sample = Column(String, primary_key=True)
    variant_id = Column(String, primary_key=True)
    chromosome = Column(String)
    position = Column(Integer)
    reference = Column(String)
    alternate = Column(String)
    quality = Column(Float)
    filter = Column(String)
    dp = Column(Integer)
    ac = Column(Integer)
    dp4_ref_forward = Column(Integer)
    dp4_ref_reverse = Column(Integer)
    dp4_alt_forward = Column(Integer)
    dp4_alt_reverse = Column(Integer)
    vaf = Column(Float)
    strand_bias = Column(Integer)

    # fields replicated from Variant for performance reasons
    annotation = Column(String)
    annotation_impact = Column(String)
    biotype = Column(String)
    annotation_highest_impact = Column(String, index=True)
    gene_name = Column(String)
    hgvs_c = Column(String)
    hgvs_p = Column(String)

    # fields replicated from sample for performance reasons
    date = Column(Date)

    # derived annotations
    variant_type = Column(Enum(VariantType, name=VariantType.__constraint_name__))
    length = Column(Integer)
    reference_amino_acid = Column(String)
    alternate_amino_acid = Column(String)
    position_amino_acid = Column(Integer)

    # ConsHMM conservation annotations
    cons_hmm_sars_cov_2 = Column(Float)
    cons_hmm_sarbecovirus = Column(Float)
    cons_hmm_vertebrate_cov = Column(Float)

    # Pfam protein domains
    pfam_name = Column(String)
    pfam_description = Column(String)

    ForeignKeyConstraint([sample], [SampleEna.run_accession])
    ForeignKeyConstraint([variant_id], [LowFrequencyVariant.variant_id])

    __table_args__ = (
        Index("{}_index_position".format(LOW_FREQUENCY_VARIANT_OBSERVATION_TABLE_NAME), "position"),
        Index("{}_index_annotation_vaf".format(LOW_FREQUENCY_VARIANT_OBSERVATION_TABLE_NAME), "annotation_highest_impact", "vaf"),
        Index("{}_index_vaf".format(LOW_FREQUENCY_VARIANT_OBSERVATION_TABLE_NAME), "vaf"),
    )


class GisaidVariantObservation(Base):
    """
    A variant observation in a particular sample. This contains all annotations of a specific observation of a variant.
    """
    __tablename__ = GISAID_VARIANT_OBSERVATION_TABLE_NAME

    sample = Column(String, primary_key=True)
    variant_id = Column(String, primary_key=True)
    chromosome = Column(String)
    position = Column(Integer)
    reference = Column(String)
    alternate = Column(String)
    quality = Column(Float)
    filter = Column(String)
    dp = Column(Integer)
    ac = Column(Integer)
    dp4_ref_forward = Column(Integer)
    dp4_ref_reverse = Column(Integer)
    dp4_alt_forward = Column(Integer)
    dp4_alt_reverse = Column(Integer)
    vaf = Column(Float)
    strand_bias = Column(Integer)

    # fields replicated from Variant for performance reasons
    annotation = Column(String)
    annotation_impact = Column(String)
    biotype = Column(String)
    annotation_highest_impact = Column(String, index=True)
    gene_name = Column(String)
    hgvs_c = Column(String)
    hgvs_p = Column(String)

    # fields replicated from sample for performance reasons
    date = Column(Date)

    # derived annotations
    variant_type = Column(Enum(VariantType, name=VariantType.__constraint_name__))
    length = Column(Integer)
    reference_amino_acid = Column(String)
    alternate_amino_acid = Column(String)
    position_amino_acid = Column(Integer)

    # ConsHMM conservation annotations
    cons_hmm_sars_cov_2 = Column(Float)
    cons_hmm_sarbecovirus = Column(Float)
    cons_hmm_vertebrate_cov = Column(Float)

    # Pfam protein domains
    pfam_name = Column(String)
    pfam_description = Column(String)

    ForeignKeyConstraint([sample], [SampleGisaid.run_accession])
    ForeignKeyConstraint([variant_id], [GisaidVariant.variant_id])

    __table_args__ = (
        Index("{}_index_annotation_position".format(GISAID_VARIANT_OBSERVATION_TABLE_NAME),
              "annotation_highest_impact", "position"),
        Index("{}_index_sample".format(GISAID_VARIANT_OBSERVATION_TABLE_NAME), "sample"),
        Index("{}_index_position".format(GISAID_VARIANT_OBSERVATION_TABLE_NAME), "position"),
        Index("{}_index_annotation".format(GISAID_VARIANT_OBSERVATION_TABLE_NAME), "annotation_highest_impact"),
        Index("{}_index_variant_id".format(GISAID_VARIANT_OBSERVATION_TABLE_NAME), "variant_id")
    )


class VariantCooccurrence(Base):

    __tablename__ = VARIANT_COOCCURRENCE_TABLE_NAME

    variant_id_one = Column(String, primary_key=True)
    variant_id_two = Column(String, primary_key=True)
    count = Column(Integer, default=0)

    ForeignKeyConstraint([variant_id_one], [Variant.variant_id])
    ForeignKeyConstraint([variant_id_two], [Variant.variant_id])


class CovigatorModule(enum.Enum):

    __constraint_name__ = COVIGATOR_MODULE_CONSTRAINT_NAME

    ACCESSOR = 1
    PROCESSOR = 2
    DOWNLOADER = 3


class Log(Base):
    """
    The table that holds an ENA job
    """
    __tablename__ = LOG_TABLE_NAME

    id = Column(Integer, primary_key=True, autoincrement=True)

    start = Column(DateTime(timezone=True), nullable=False)
    end = Column(DateTime(timezone=True), nullable=False)
    source = Column(Enum(DataSource, name=DataSource.__constraint_name__), nullable=False)
    module = Column(Enum(CovigatorModule, name=CovigatorModule.__constraint_name__), nullable=False)

    has_error = Column(Boolean, default=False)
    error_message = Column(String)
    processed = Column(Integer)
    data = Column(JSON)


class LastUpdate(Base):

    __tablename__ = LAST_UPDATE_TABLE_NAME

    id = Column(Integer, primary_key=True, autoincrement=True)
    update_time = Column(Date(), nullable=False)
    source = Column(Enum(DataSource, name=DataSource.__constraint_name__), nullable=False)


class Conservation(Base):
    """
    Table to hold conservation scores
    """
    __tablename__ = CONSERVATION_TABLE_NAME

    chromosome = Column(String, primary_key=True)
    start = Column(Integer, primary_key=True)
    end = Column(Integer, primary_key=True)
    conservation = Column(Float)
    conservation_sarbecovirus = Column(Float)
    conservation_vertebrates = Column(Float)


class PrecomputedVariantsPerSample(Base):

    __tablename__ = PRECOMPUTED_VARIANTS_PER_SAMPLE_TABLE_NAME

    id = Column(Integer, primary_key=True, autoincrement=True)
    count = Column(Integer)
    number_mutations = Column(Integer)
    source = Column(Enum(DataSource, name=DataSource.__constraint_name__))
    variant_type = Column(Enum(VariantType, name=VariantType.__constraint_name__))
    gene_name = Column(String)


class PrecomputedSubstitutionsCounts(Base):

    __tablename__ = PRECOMPUTED_SUBSTITUTIONS_COUNTS_TABLE_NAME

    id = Column(Integer, primary_key=True, autoincrement=True)
    count = Column(Integer)
    reference = Column(String)
    alternate = Column(String)
    source = Column(Enum(DataSource, name=DataSource.__constraint_name__))
    variant_type = Column(Enum(VariantType, name=VariantType.__constraint_name__))
    gene_name = Column(String)


class PrecomputedIndelLength(Base):

    __tablename__ = PRECOMPUTED_INDEL_LENGTH_TABLE_NAME

    id = Column(Integer, primary_key=True, autoincrement=True)
    count = Column(Integer)
    length = Column(Integer)
    source = Column(Enum(DataSource, name=DataSource.__constraint_name__))
    gene_name = Column(String)


class PrecomputedAnnotation(Base):

    __tablename__ = PRECOMPUTED_ANNOTATION_TABLE_NAME

    id = Column(Integer, primary_key=True, autoincrement=True)
    count = Column(Integer)
    annotation = Column(String)
    source = Column(Enum(DataSource, name=DataSource.__constraint_name__))
    gene_name = Column(String)


class PrecomputedOccurrence(Base):

    __tablename__ = PRECOMPUTED_OCCURRENCE_TABLE_NAME

    id = Column(Integer, primary_key=True, autoincrement=True)
    total = Column(Integer)
    count = Column(Integer)
    month = Column(String)
    frequency = Column(Float)
    frequency_by_month = Column(Float)
    variant_id = Column(String)
    hgvs_p = Column(String)
    gene_name = Column(String)
    domain = Column(String)
    annotation = Column(String)
    source = Column(Enum(DataSource, name=DataSource.__constraint_name__))


class RegionType(enum.Enum):

    __constraint_name__ = REGION_TYPE_CONSTRAINT_NAME

    GENE = 1
    DOMAIN = 2
    CODING_REGION=3


class PrecomputedSynonymousNonSynonymousCounts(Base):

    __tablename__ = PRECOMPUTED_NS_S_COUNTS_TABLE_NAME

    id = Column(Integer, primary_key=True, autoincrement=True)
    month = Column(Date)
    region_type = Column(Enum(RegionType, name=RegionType.__constraint_name__))
    region_name = Column(String)
    country = Column(String)
    source = Column(Enum(DataSource, name=DataSource.__constraint_name__))
    ns = Column(Integer)
    s = Column(Integer)


class PrecomputedTableCounts(Base):

    __tablename__ = PRECOMPUTED_TABLE_COUNTS_TABLE_NAME
    VIRTUAL_TABLE_COUNTRY = "Country"
    FACTOR_SOURCE = "source"

    id = Column(Integer, primary_key=True, autoincrement=True)
    table = Column(String)
    factor = Column(String)
    value = Column(String)
    count = Column(Integer)


class PrecomputedVariantAbundanceHistogram(Base):

    __tablename__ = PRECOMPUTED_VARIANT_ABUNDANCE_HIST_TABLE_NAME

    id = Column(Integer, primary_key=True, autoincrement=True)
    position_bin = Column(Integer)
    count_unique_variants = Column(Integer)
    count_variant_observations = Column(Integer)
    bin_size = Column(Integer)
    source = Column(Enum(DataSource, name=DataSource.__constraint_name__))


class PrecomputedVariantsPerLineage(Base):

    __tablename__ = PRECOMPUTED_VARIANTS_PER_LINEAGE_TABLE_NAME

    id = Column(Integer, primary_key=True, autoincrement=True)
    lineage = Column(String)
    variant_id = Column(String)
    country = Column(String)
    count_observations = Column(Integer)
    source = Column(Enum(DataSource, name=DataSource.__constraint_name__))



import numpy as np
from itertools import combinations
from typing import Tuple, Union
from faker import Faker
from sqlalchemy.orm import Session

from covigator import MISSENSE_VARIANT, SYNONYMOUS_VARIANT, INFRAME_INSERTION, INFRAME_DELETION
from covigator.database.model import SampleEna, DataSource, JobStatus, Log, CovigatorModule, Variant, \
    VariantCooccurrence, VariantType, SampleGisaid, GisaidVariant
from Bio.Alphabet.IUPAC import IUPACData

from covigator.database.queries import Queries

MOCKED_GENES = ["S", "N", "E", "M", "ORF3a", "ORF1ab", "ORF7b", "ORF10", "ORF6", "ORF8", "ORF7a"]
MOCKED_DOMAINS = ["CoV_S2", "bCoV_S1_RBD", "bCoV_S1_N", "CoV_S1_C"]
MOCKED_LINEAGES = ["A", "B", "C", "D", "E", "F", ""]


def get_mocked_variant(faker: Faker, chromosome=None, gene_name=None, source=DataSource.ENA.name, session: Session = None) -> Variant:

    if gene_name is None:
        gene_name = faker.random_choices(MOCKED_GENES, length=1)[0]
    domain_name = faker.random_choices(MOCKED_DOMAINS, length=1)[0]
    annotation = faker.random_choices(
            [MISSENSE_VARIANT, SYNONYMOUS_VARIANT, INFRAME_DELETION, INFRAME_INSERTION], length=1)[0]

    if session:
        queries = Queries(session=session)
        gene = queries.get_gene(gene_name=gene_name)
        start = gene.start
        end = gene.end
    else:
        start = 1
        end = 30000

    klass = Queries.get_variant_klass(source)
    variant = klass(
        chromosome=chromosome if chromosome else faker.bothify(text="chr##"),
        position=faker.random_int(min=start, max=end),
        reference=faker.random_choices(list(IUPACData.unambiguous_dna_letters), length=1)[0],
        # TODO: reference and alternate could be equal!
        alternate=faker.random_choices(list(IUPACData.unambiguous_dna_letters), length=1)[0],
        variant_type=VariantType.SNV,
        gene_name=gene_name,
        hgvs_p="p.{}{}{}".format(
            faker.random_choices(list(IUPACData.protein_letters_1to3.values()), length=1)[0],
            faker.random_int(min=1, max=500),
            faker.random_choices(list(IUPACData.protein_letters_1to3.values()), length=1)[0]
        ),
        annotation=annotation,
        annotation_highest_impact=annotation,
        pfam_name=domain_name
    )
    variant.variant_id = variant.get_variant_id()
    return variant


def get_mocked_variant_observation(
        sample: Union[SampleEna, SampleGisaid], variant: Union[Variant or GisaidVariant], faker=Faker()):

    klass = Queries.get_variant_observation_klass(
        DataSource.ENA.name if isinstance(sample, SampleEna) else DataSource.GISAID.name)
    return klass(
        sample=sample.run_accession if sample else faker.unique.uuid4(),
        variant_id=variant.variant_id,
        chromosome=variant.chromosome,
        position=variant.position,
        reference=variant.reference,
        alternate=variant.alternate,
        variant_type=VariantType.SNV,
        annotation=variant.annotation,
        annotation_highest_impact=variant.annotation_highest_impact,
        gene_name=variant.gene_name,
        pfam_name=variant.pfam_name,
        date=faker.date_time(),
        hgvs_p=variant.hgvs_p
    )


def get_mocked_sample(faker: Faker, source: DataSource, job_status=JobStatus.FINISHED) -> Union[SampleEna, SampleGisaid]:
    identifier = faker.unique.uuid4()
    if source == DataSource.ENA:
        sample = SampleEna(
            run_accession=identifier,
            collection_date=faker.date_time(),
            country=faker.country(),
            fastq_ftp=faker.uri(),
            fastq_md5=faker.md5(),
            num_fastqs=1,
            status=job_status,
            pangolin_lineage=faker.random_choices(MOCKED_LINEAGES, length=1)[0]
        )
    elif source == DataSource.GISAID:
        sample = SampleGisaid(
            run_accession=identifier,
            collection_date=faker.date_time(),
            country=faker.country(),
            status=job_status,
            pangolin_lineage=faker.random_choices(MOCKED_LINEAGES, length=1)[0]
        )
    else:
        raise ValueError("Bad data source")
    return sample


def get_mocked_ena_sample(faker: Faker, job_status=JobStatus.FINISHED) -> SampleEna:
    return get_mocked_sample(faker=faker, source=DataSource.ENA, job_status=job_status)


def get_mocked_gisaid_sample(faker: Faker, job_status=JobStatus.FINISHED) -> SampleGisaid:
    return get_mocked_sample(faker=faker, source=DataSource.GISAID, job_status=job_status)


def get_mocked_log(faker: Faker, source: DataSource = None, module: CovigatorModule = None) -> Log:
    return Log(
        start=faker.date_time(),
        end=faker.date_time(),
        source=source if source else faker.random_choices((DataSource.ENA, DataSource.GISAID), length=1)[0],
        module=module if module else faker.random_choices((CovigatorModule.ACCESSOR, CovigatorModule.PROCESSOR),
                                                          length=1)[0],
        has_error=faker.boolean(),
        processed=faker.random_digit(),
        data={"included": faker.random_digit(),
              "excluded": {"this": faker.random_digit(), "that": faker.random_digit()}}
    )


def get_mocked_variant_cooccurrence(faker: Faker, variant_one: Variant, variant_two: Variant) -> VariantCooccurrence:
    if variant_one.position <= variant_two.position:
        cooccurrence = VariantCooccurrence(
            variant_id_one=variant_one.variant_id,
            variant_id_two=variant_two.variant_id,
            count=faker.random_int(min=1, max=10)
        )
    else:
        cooccurrence = VariantCooccurrence(
            variant_id_two=variant_one.variant_id,
            variant_id_one=variant_two.variant_id,
            count=faker.random_int(min=1, max=10)
        )
    return cooccurrence


def mock_samples_and_variants(faker, session: Session, num_samples=10):
    existing_variants = {DataSource.ENA.name: set(), DataSource.GISAID.name: set()}
    samples = mock_samples(faker=faker, session=session, num_samples=num_samples)
    for sample in samples:
        source = DataSource.ENA if isinstance(sample, SampleEna) else DataSource.GISAID
        variants = [get_mocked_variant(faker=faker, source=source.name, session=session) for _ in range(10)]
        # this aims at removing potentially repeated variants
        variants_dict = {v.variant_id: v for v in variants}
        variants = variants_dict.values()
        new_variants = list(filter(lambda x: x.variant_id not in existing_variants.get(source.name), variants))
        session.add_all(new_variants)
        session.commit()
        existing_variants.get(source.name).update([v.variant_id for v in variants])

        variants_observations = [get_mocked_variant_observation(faker=faker, variant=v, sample=sample)
                                 for v in variants]
        session.add_all(variants_observations)
        session.commit()


def mock_samples(faker, session: Session, num_samples=10, job_status=JobStatus.FINISHED, source=None):
    samples = []
    for _ in range(num_samples):
        if source is not None:
            selected_source = source
        else:
            selected_source = faker.random_choices([DataSource.ENA.name, DataSource.GISAID.name], length=1)[0]
        if selected_source == DataSource.ENA.name:
            sample= get_mocked_ena_sample(faker=faker, job_status=job_status)
        else:   # GISAID
            sample = get_mocked_gisaid_sample(faker=faker, job_status=job_status)
        samples.append(sample)

    session.add_all(samples)
    session.commit()
    return samples


def mock_cooccurrence_matrix(faker, session: Session):
    # add some variants belonging to two genes
    chromosome = "fixed_chromosome"
    gene_name = "S"
    variants = [get_mocked_variant(faker=faker, chromosome=chromosome, gene_name=gene_name) for _ in range(5)]
    other_gene_name = "N"
    other_variants = [get_mocked_variant(faker=faker, chromosome=chromosome, gene_name=other_gene_name) for _
                      in range(5)]
    session.add_all(variants + other_variants)
    session.commit()
    # adds some cooccurrences
    cooccurrences = []
    other_cooccurrences = []
    variants_to_sample = {"{}-{}".format(v1.hgvs_p, v2.hgvs_p): (v1, v2) for v1, v2 in
                          list(combinations(variants, 2))}
    other_variants_to_sample = {"{}-{}".format(v1.hgvs_p, v2.hgvs_p): (v1, v2) for v1, v2 in
                                list(combinations(other_variants, 2))}
    combined_variants_to_sample = {"{}-{}".format(v1.hgvs_p, v2.hgvs_p): (v1, v2) for v1, v2 in
                                   list(zip(variants, other_variants))}
    for variant in variants + other_variants:
        cooccurrences.append(get_mocked_variant_cooccurrence(faker, variant, variant))
    for (variant_one, variant_two) in [variants_to_sample.get(k) for k in
                                       np.random.choice(list(variants_to_sample.keys()), 5, replace=False)]:
        cooccurrences.append(get_mocked_variant_cooccurrence(faker, variant_one, variant_two))
    for (variant_one, variant_two) in [other_variants_to_sample.get(k) for k in
                                       np.random.choice(list(other_variants_to_sample.keys()), 5, replace=False)]:
        other_cooccurrences.append(get_mocked_variant_cooccurrence(faker, variant_one, variant_two))
    for (variant_one, variant_two) in [combined_variants_to_sample.get(k) for k in
                                       np.random.choice(list(combined_variants_to_sample.keys()), 5,
                                                        replace=False)]:
        other_cooccurrences.append(get_mocked_variant_cooccurrence(faker, variant_one, variant_two))
    session.add_all(cooccurrences + other_cooccurrences)
    session.commit()

    # add some samples to compute the frequency right
    mock_samples(faker=faker, session=session, num_samples=10)

    return other_variants, variants



import pandas as pd
from logzero import logger
from sqlalchemy import func, desc
from sqlalchemy.orm import Session
from covigator import SYNONYMOUS_VARIANT
from covigator.database.model import DataSource, PrecomputedOccurrence
from covigator.database.queries import Queries


NUMBER_TOP_OCCURRENCES = 1000


class TopOccurrencesLoader:

    def __init__(self, session: Session):
        self.session = session
        self.queries = Queries(session=self.session)

    def load(self):

        database_rows = self._read_top_occurrent_variants()

        # delete all rows before starting
        self.session.query(PrecomputedOccurrence).delete()
        self.session.commit()

        if len(database_rows) > 0:
            self.session.add_all(database_rows)
            self.session.commit()
        logger.info("Added {} entries to {}".format(len(database_rows), PrecomputedOccurrence.__tablename__))

    def _read_top_occurrent_variants(self):
        # gets the top occurrent variants for each source
        top_occurring_variants_ena = None
        try:
            top_occurring_variants_ena = self.get_top_occurring_variants(
                top=NUMBER_TOP_OCCURRENCES, source=DataSource.ENA.name)
        except ValueError as e:
            logger.exception(e)
            logger.error("No top occurrences for ENA data")
        top_occurring_variants_gisaid = None
        try:
            top_occurring_variants_gisaid = self.get_top_occurring_variants(
                top=NUMBER_TOP_OCCURRENCES, source=DataSource.GISAID.name)
        except ValueError:
            logger.error("No top occurrences for GISAID data")
        database_rows = []
        # stores the precomputed data
        if top_occurring_variants_ena is not None:
            for index, row in top_occurring_variants_ena.iterrows():
                # add entries per gene
                database_rows.append(self._row_to_top_occurrence(row, source=DataSource.ENA))
        if top_occurring_variants_gisaid is not None:
            for index, row in top_occurring_variants_gisaid.iterrows():
                # add entries per gene
                database_rows.append(self._row_to_top_occurrence(row, source=DataSource.GISAID))
        return database_rows

    def _row_to_top_occurrence(self, row, source):
        return PrecomputedOccurrence(
            total=row["total"],
            frequency=row["frequency"],
            variant_id=row["variant_id"],
            hgvs_p=row["hgvs_p"],
            gene_name=row["gene_name"],
            domain=row["pfam_name"],
            annotation=row["annotation_highest_impact"],
            source=source,
            month=row["month"],
            count=row["count"],
            frequency_by_month=row["frequency_by_month"],
        )

    def get_top_occurring_variants(self, top, source: str):
        klass = self.queries.get_variant_observation_klass(source)
        query = self.session.query(
            klass.variant_id, klass.hgvs_p, klass.gene_name, klass.pfam_name,
            klass.annotation_highest_impact, func.count().label('total')) \
            .filter(klass.annotation_highest_impact != SYNONYMOUS_VARIANT)
        query = query.group_by(klass.variant_id, klass.hgvs_p, klass.gene_name,
                      klass.pfam_name, klass.annotation_highest_impact) \
            .order_by(desc('total')).limit(top)
        top_occurring_variants = pd.read_sql(query.statement, self.session.bind)

        # calculate frequency
        count_samples = self.queries.count_samples(source=source)
        top_occurring_variants['frequency'] = top_occurring_variants.total.transform(
            lambda x: round(float(x) / count_samples, 3))

        # add counts for every month
        top_occurring_variants = self._get_counts_per_month(
            top_occurring_variants=top_occurring_variants, source=source)

        return top_occurring_variants

    def _get_counts_per_month(self, top_occurring_variants, source: str):
        variant_counts_by_month = []
        for _, variant in top_occurring_variants.iterrows():
            variant_counts_by_month.append(self.queries.get_variant_counts_by_month(variant.variant_id, source=source))
        if len(variant_counts_by_month) > 1:
            top_occurring_variants_by_month = pd.concat(variant_counts_by_month)
            # get total count of samples per month to calculate the frequency by month
            sample_counts_by_month = self.queries.get_sample_counts_by_month(source=source)
            top_occurring_variants_by_month = pd.merge(
                left=top_occurring_variants_by_month, right=sample_counts_by_month, how="left", on="month")
            top_occurring_variants_by_month["frequency_by_month"] = \
                (top_occurring_variants_by_month["count"] / top_occurring_variants_by_month["sample_count"]). \
                    transform(lambda x: round(x, 3))
            # join both tables with total counts and counts per month
            top_occurring_variants = pd.merge(
                left=top_occurring_variants, right=top_occurring_variants_by_month, on="variant_id", how="left")
            # format the month column appropriately
            top_occurring_variants.month = top_occurring_variants.month.transform(
                lambda d: "{}-{:02d}".format(d.year, int(d.month)))
        return top_occurring_variants



import traceback
from contextlib import suppress
from datetime import datetime
from logzero import logger
from covigator.configuration import Configuration
from covigator.database.database import Database
from covigator.database.model import Log, DataSource, CovigatorModule, JobStatus
from covigator.database.queries import Queries
from covigator.misc import backoff_retrier
from covigator.pipeline.downloader import Downloader


class EnaDownloader:

    def __init__(self, database: Database, config: Configuration, wait_time=60):
        self.config = config
        self.start_time = datetime.now()
        self.has_error = False
        self.error_message = None
        self.database = database
        assert self.database is not None, "Empty database"
        self.session = self.database.get_database_session()
        self.queries = Queries(self.session)
        self.wait_time = wait_time
        self.downloader = Downloader(config=config)

    def process(self):
        logger.info("Starting ENA downloader")
        count = 0
        try:
            while True:
                # queries 100 jobs every time to make sending to queue faster
                samples = self.queries.find_first_jobs_to_download(DataSource.ENA, n=1000)
                if samples is None or len(samples) == 0:
                    logger.info("No more ENA samples to download after downloading {} samples".format(count))
                    break
                for sample in samples:
                    # downloads
                    try:
                        download_with_retries = backoff_retrier.wrapper(self.downloader.download, self.config.retries_download)
                        fastq_path = download_with_retries(sample_ena=sample)
                        sample.fastq_path = fastq_path
                        sample.downloaded_at = datetime.now()
                        sample.sample_folder = sample.get_sample_folder(base_folder=self.config.storage_folder)
                        sample.status = JobStatus.DOWNLOADED
                        self.session.commit()
                    except Exception as e:
                        sample.status = JobStatus.FAILED_DOWNLOAD
                        sample.error_message = self._get_traceback_from_exception(e)
                        self.session.commit()

                    count += 1
                    if count % 1000 == 0:
                        logger.info("Downloaded {} ENA samples...".format(count))

            logger.info("ENA downloader finished!")

        except Exception as e:
            logger.exception(e)
            self.session.rollback()
            self.has_error = True
        finally:
            logger.info("Logging execution stats...")
            self._write_execution_log(count, data_source=DataSource.ENA)
            with suppress(Exception):
                self.session.close()
            logger.info("Database sessions closed")

    def _write_execution_log(self, count, data_source: DataSource):
        end_time = datetime.now()
        self.session.add(Log(
            start=self.start_time,
            end=end_time,
            source=data_source,
            module=CovigatorModule.DOWNLOADER,
            has_error=self.has_error,
            processed=count,
            data={
                "processed": count
            }
        ))
        self.session.commit()

    @staticmethod
    def _get_traceback_from_exception(e):
        return "".join(traceback.format_exception(
            etype=type(e), value=e, tb=e.__traceback__))




from typing import List
import numpy as np
import plotly
from logzero import logger

from covigator import MISSENSE_VARIANT, SYNONYMOUS_VARIANT
from covigator.dashboard.figures import VARIANT_TYPE_COLOR_MAP
from covigator.dashboard.figures.figures import Figures, PLOTLY_CONFIG, MARGIN, TEMPLATE
import plotly.express as px
import plotly.graph_objects as go
import dash_core_components as dcc


INDEL_TYPE_COLOR_MAP = {
    "INSERTION_INFRAME": "#ee6002",
    "DELETION_INFRAME": "#09af00",
    "INSERTION_FRAMESHIFT": "#ffddb0",
    "DELETION_FRAMESHIFT": "#defabb",
}


class MutationStatsFigures(Figures):

    def get_annotations_plot(self, data_source: str = None, genes: List[str] = None):
        logger.debug("Getting data on annotations...")
        data = self.queries.get_annotations(data_source=data_source, genes=genes)
        graph = dcc.Markdown("""**No data for the current selection**""")
        if data is not None and data.shape[0] > 0:
            logger.debug("Prepare plot on annotations...")
            fig = px.bar(data, y="count", x="annotation", log_y=True, color='count',
                         color_continuous_scale=plotly.colors.sequential.Brwnyl)
            #.update_yaxes(categoryorder="total descending")
            fig.update_layout(
                margin=MARGIN,
                template=TEMPLATE,
                legend={'traceorder': 'normal', 'title': None},
                yaxis={'title': "num. mutations (log)"},  # , 'autorange': 'reversed'},
                xaxis={'title': None},
            )
            graph = [
                dcc.Graph(figure=fig, config=PLOTLY_CONFIG),
                dcc.Markdown("""
                **Most common mutation effects**
                
                *Ratio of non synonymous to synonymous SNVs (N/S): {dnds}*
                """.format(dnds=round(data[data.annotation == MISSENSE_VARIANT]["count"].sum() /
                                data[data.annotation == SYNONYMOUS_VARIANT]["count"].sum(), 3)))
            ]
        return graph

    def get_indels_lengths_plot(self, data_source: str = None, genes: List[str] = None):
        logger.debug("Getting data on indels lengths...")
        data = self.queries.get_indel_lengths(data_source=data_source, genes=genes)
        graph = dcc.Markdown("""**No data for the current selection**""")
        if data is not None and data.shape[0] > 0:
            logger.debug("Prepare plot on indels lengths...")
            fig = px.bar(
                data, y="count", x="length", color="variant_type", color_discrete_map=INDEL_TYPE_COLOR_MAP)
            fig.update_layout(
                margin=MARGIN,
                template=TEMPLATE,
                legend={'traceorder': 'normal', 'title': None},
                yaxis={'title': "num. samples"},
                xaxis={'title': "indel length (bp)"}
            )
            graph = [
                dcc.Graph(figure=fig, config=PLOTLY_CONFIG),
                dcc.Markdown("""
                **Indel length distribution**
                
                *Insertion to deletion ratio: {indel_ratio}*
                """.format(
                    indel_ratio=round(data[
                        data.variant_type.isin(["INSERTION_INFRAME", "INSERTION_FRAMESHIFT"])]["count"].sum() /
                        data[data.variant_type.isin(["DELETION_INFRAME", "DELETION_FRAMESHIFT"])]["count"].sum(), 3)))
            ]
        return graph

    def get_substitutions_plot(self, data_source: str, variant_types: List[str], genes: List[str] = None):
        logger.debug("Getting data on substitutions plot...")
        data = self.queries.get_substitutions(data_source=data_source, genes=genes, variant_types=variant_types)
        graph = dcc.Markdown("""**No data for the current selection**""")
        if data is not None and data.shape[0] > 0:
            logger.debug("Preparing plot on samples by country...")
            fig = px.bar(
                data, y="substitution", x="count", color="variant_type", text="rate",
                color_discrete_map=VARIANT_TYPE_COLOR_MAP)\
                .update_yaxes(categoryorder="total descending")
            fig.update_traces(textposition='outside')
            fig.update_layout(
                margin=go.layout.Margin(l=0, r=40, b=0, t=30),  # we need some extra space for some labels overflowing
                template=TEMPLATE,
                legend={'title': None, 'yanchor': "bottom", 'y': 0.01, 'xanchor': "right", 'x': 0.99},
                showlegend=True,
                yaxis={'title': None, 'autorange': 'reversed'},
                xaxis={'title': "num. samples"},
                uniformtext_mode='show',
                uniformtext_minsize=8
            )
            graph = [
                dcc.Graph(figure=fig, config=PLOTLY_CONFIG),
                dcc.Markdown("""
                **Top mutations**
                
                *Only mutations occurring at least in 10 samples are represented*
                """)
            ]
        return graph

    def get_variants_per_sample_plot(
            self, data_source: str, genes: List[str] = None, variant_types: List[str] = None):

        logger.debug("Getting data on variants per sample...")
        data = self.queries.get_variants_per_sample(data_source=data_source, genes=genes, variant_types=variant_types)
        graph = dcc.Markdown("""**No data for the current selection**""")
        if data is not None and data.shape[0] > 0:
            logger.debug("Preparing plot on samples by country...")
            counts = np.repeat(data.number_mutations, data["count"])
            median = round(np.median(counts), 3)
            third_quartile = np.percentile(counts, 75)
            first_quartile = np.percentile(counts, 25)
            extreme_threshold = median + (3 * (third_quartile - first_quartile))
            filtered_data = data[data.number_mutations <= extreme_threshold]
            if filtered_data.shape[0] > 0:
                fig = px.bar(filtered_data,
                             x="number_mutations",
                             y='count',
                             color="variant_type",
                             color_discrete_map=VARIANT_TYPE_COLOR_MAP)
                fig.add_vline(x=median, line_width=2, line_dash="dash", line_color='grey',
                              annotation_text="median", annotation_position="top right")
                fig.add_vrect(x0=first_quartile, x1=third_quartile,
                              fillcolor="grey", opacity=0.25, line_width=0)
                fig.update_layout(
                    margin=MARGIN,
                    template=TEMPLATE,
                    legend={'traceorder': 'normal', 'title': None},
                    yaxis={'title': "num. samples"},
                    xaxis={'title': "num. mutations"},
                )

                graph = [
                    dcc.Graph(figure=fig, config=PLOTLY_CONFIG),
                    dcc.Markdown("""
                            **Mutations per sample**
    
                            *Median: {median} (IQR: {iqr})*
                            """.format(median=median, iqr=third_quartile - first_quartile))
                ]
        return graph



from unittest import TestCase

from covigator.configuration import Configuration
from covigator.database.model import SampleEna
from covigator.pipeline.downloader import Downloader
import os


class DownloaderTest(TestCase):

    def setUp(self) -> None:
        self.downloader = Downloader(config=Configuration())

    def test_download_ena_run(self):
        run_accession = "TEST12345"
        ena_run = SampleEna(
            run_accession=run_accession,
            fastq_ftp="ftp.sra.ebi.ac.uk/vol1/fastq/ERR419/002/ERR4192722/ERR4192722.fastq.gz;"
                       "ftp.sra.ebi.ac.uk/vol1/fastq/ERR419/002/ERR4192722/ERR4192722_1.fastq.gz;"
                       "ftp.sra.ebi.ac.uk/vol1/fastq/ERR419/002/ERR4192722/ERR4192722_2.fastq.gz",
            fastq_md5="f310c1f2952f07d70aee7464a276cba1;"
                      "11a6108b741b48402001e0bbae16cbaa;"
                      "2dc39ea2a383550c15013435b65489f8"
        )
        self.downloader.download(ena_run)
        run_storage_folder = os.path.join(self.downloader.storage_folder, run_accession)
        self.assertTrue(os.path.exists(run_storage_folder))
        self.assertTrue(os.path.exists(os.path.join(run_storage_folder, "ERR4192722.fastq.gz")))
        self.assertTrue(os.path.exists(os.path.join(run_storage_folder, "ERR4192722_1.fastq.gz")))
        self.assertTrue(os.path.exists(os.path.join(run_storage_folder, "ERR4192722_2.fastq.gz")))



from typing import List
import numpy as np
import pandas as pd
from logzero import logger
from covigator.dashboard.figures.figures import Figures, PLOTLY_CONFIG, MARGIN, TEMPLATE
import plotly.express as px
import dash_core_components as dcc

from covigator.database.model import DataSource


class SampleFigures(Figures):

    def get_accumulated_samples_by_country_plot(self, data_source: str, countries=None, min_samples=1000,
                                                lineages=None):
        logger.debug("Getting data on samples by country...")
        data = self.queries.get_accumulated_samples_by_country(
            data_source=data_source, countries=countries, min_samples=min_samples, lineages=lineages)
        graph = dcc.Markdown("""**No data for the current selection**""")
        if data is not None and data.shape[0] > 0:
            logger.debug("Prepare plot on samples by country...")
            countries = list(data.sort_values("cumsum", ascending=False).country.unique())
            fig = px.area(data, x="date", y="cumsum", color="country",
                          category_orders={
                              "country": countries[::-1]},
                          labels={"cumsum": "num. samples", "count": "increment"},
                          hover_data=["count"],
                          color_discrete_sequence=px.colors.qualitative.Vivid)
            fig.update_traces(line=dict(width=0.5))
            fig.update_layout(
                margin=MARGIN,
                template=TEMPLATE,
                legend={'traceorder': 'reversed', 'title': None},
                xaxis={'title': None},
            )

            top_countries = countries[0: min(5, len(countries))]
            top_countries_and_cumsum = data[data.country.isin(top_countries)][["country", "cumsum"]] \
                .groupby("country").max().reset_index().sort_values("cumsum", ascending=False)
            top_countries_tooltip = list(
                top_countries_and_cumsum.apply(lambda x: "{} ({})".format(x[0], int(x[1])), axis=1))

            graph = [
                dcc.Graph(figure=fig, config=PLOTLY_CONFIG),
                dcc.Markdown("""
                **Accumulated samples by country**

                *Top {} countries: {}. Countries with < {} samples are excluded*
                """.format(len(top_countries_tooltip),
                           ", ".join(top_countries_tooltip),
                           min_samples))
            ]
        return graph

    def _calculate_dn_ds(self, NS, S, ns, s):
        pn = float(ns) / float(NS)
        ps = float(s) / float(S)
        dn = np.log(1 + pn) + 1
        ds = np.log(1 + ps) + 1
        return dn / ds

    def get_dnds_by_gene_plot(
            self, data_source: str, countries: List[str] =None, genes: List[str] = None):

        logger.debug("Getting data on dN/dS...")
        data = self.queries.get_dnds_table(
            source=data_source, countries=countries, genes=genes)
        graph = dcc.Markdown("""**No data for the current selection**""")
        if data is not None and data.shape[0] > 0:
            logger.debug("Prepare plot on dN/dS...")
            genes = pd.concat([
                self.queries.get_genes_df(),
                self.queries.get_domains_df()
            ])
            # prepares the data and calculates the dN/dS
            data_to_plot = data.groupby(["month", "region_name"]).sum().reset_index().sort_values("month")
            data_to_plot = pd.merge(left=genes, right=data_to_plot, left_on="name", right_on="region_name")
            data_to_plot["dn_ds"] = data_to_plot[["ns", "s", "fraction_non_synonymous", "fraction_synonymous"]].apply(
                lambda x: self._calculate_dn_ds(ns=x[0], s=x[1], NS=x[2], S=x[3]), axis=1)

            fig = px.line(data_to_plot, x='month', y='dn_ds', color='region_name',
                          symbol='region_name', line_dash='region_name', line_dash_sequence=['dash'],
                          labels={"dn_ds": "dN/dS", "region_name": "gene"},
                          hover_data=["region_name", "dn_ds"],
                          color_discrete_sequence=px.colors.qualitative.Vivid)
            fig.update_traces(line=dict(width=0.5), marker=dict(size=10))
            fig.update_layout(
                margin=MARGIN,
                template=TEMPLATE,
                legend={'title': None},
                xaxis={'title': None},
            )

            graph = [
                dcc.Graph(figure=fig, config=PLOTLY_CONFIG),
                dcc.Markdown("""
                **dN/dS by gene**
                """)
            ]
        return graph



import os
import json

import pandas as pd
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import Session
from covigator.database.model import Gene, Domain
from logzero import logger


class GeneAnnotationsLoader:

    GENES_NS_S_COUNTS_FILENAME = "genes_NS_S.csv"
    DOMAINS_NS_S_COUNTS_FILENAME = "domains_NS_S.csv"
    GENE_ANNOTATIONS_FILENAME = "sars_cov_2.json"

    def __init__(self, session: Session):
        self.session = session

        self.genes_ns_s_counts = os.path.join(
            os.path.abspath(os.path.dirname(__file__)), self.GENES_NS_S_COUNTS_FILENAME)
        self.domains_ns_s_counts = os.path.join(
            os.path.abspath(os.path.dirname(__file__)), self.DOMAINS_NS_S_COUNTS_FILENAME)
        self.gene_annotations = os.path.join(
            os.path.abspath(os.path.dirname(__file__)), self.GENE_ANNOTATIONS_FILENAME)

    def _remove_duplicated_genes(self, data):
        results = {}
        for g in data["genes"]:
            if g["name"] not in results:
                results[g["name"]] = g
            else:
                if int(g["end"]) - int(g["start"]) > int(results[g["name"]]["end"]) - int(results[g["name"]]["start"]):
                    # in case of duplications it keeps only the longer gene
                    results[g["name"]] = g
        data["genes"] = list(results.values())
        return data

    def load_data(self):
        # reads the JSON
        with open(self.gene_annotations) as fd:
            data = json.load(fd)

        data = self._remove_duplicated_genes(data)

        # NOTE: load fraction of synonymous and non synonymous
        genes_fractions = pd.read_csv(self.genes_ns_s_counts)
        domains_fractions = pd.read_csv(self.domains_ns_s_counts)

        count_genes = 0
        count_domains = 0

        for g in data["genes"]:
            # persist a gene
            transcript_id = g["transcripts"][0]["id"]
            gene = Gene(
                identifier=g["id"],
                name=g["name"],
                start=int(g["start"]),
                end=int(g["end"]),
                fraction_synonymous=genes_fractions[genes_fractions.transcript == transcript_id].S.iloc[0],
                fraction_non_synonymous=genes_fractions[genes_fractions.transcript == transcript_id].NS.iloc[0])
            self.session.add(gene)
            self.session.commit()
            count_genes += 1

            for d in g.get("transcripts", [])[0].get("translations", [])[0].get("protein_features"):
                if d.get("dbname") == "Pfam":
                    try:
                        domain = Domain(
                            name=d["description"],
                            description=d["interpro_description"],
                            start=int(d["start"]),
                            end=int(d["end"]),
                            fraction_synonymous=domains_fractions[domains_fractions.domain == d["description"]].S.iloc[0],
                            fraction_non_synonymous=domains_fractions[domains_fractions.domain == d["description"]].NS.iloc[0],
                            gene_identifier=gene.identifier,
                            gene_name=gene.name)
                        self.session.add(domain)
                        self.session.commit()
                    except IntegrityError:
                        logger.warn("Domain {} is duplicated in input with coordinates [{}, {}]".format(
                            d["description"], d["start"], d["end"]))
                        self.session.rollback()
                    count_domains += 1

        logger.info("Loaded into the database {} genes and {} domains".format(count_genes, count_domains))



import pandas as pd
from sqlalchemy.orm import Session
from covigator.dashboard.tabs.recurrent_mutations import BIN_SIZE_VALUES
from covigator.database.model import PrecomputedVariantsPerSample, PrecomputedSubstitutionsCounts, \
    PrecomputedIndelLength, PrecomputedAnnotation, VariantObservation, DataSource, \
    PrecomputedTableCounts, Variant, SubclonalVariantObservation, PrecomputedVariantAbundanceHistogram, \
    SampleEna, SampleGisaid, GisaidVariantObservation, GisaidVariant
from logzero import logger

from covigator.database.queries import Queries
from covigator.precomputations.load_ns_s_counts import NsSCountsLoader
from covigator.precomputations.load_top_occurrences import TopOccurrencesLoader
from covigator.precomputations.load_variants_per_lineage import VariantsPerLineageLoader


class PrecomputationsLoader:

    def __init__(self, session: Session):
        self.session = session
        self.queries = Queries(session=self.session)
        self.ns_s_counts_loader = NsSCountsLoader(session=session)
        self.top_occurrences_loader = TopOccurrencesLoader(session=session)
        self.variants_per_lineage_loader = VariantsPerLineageLoader(session=session)

    def load(self):
        logger.info("Starting precomputations...")
        self.load_table_counts()
        logger.info("Done with table counts (1/9)")
        self.load_variant_abundance_histogram()
        logger.info("Done with variant abundance histogram (2/9)")
        self.load_counts_variants_per_sample()
        logger.info("Done with count variants per sample (3/9)")
        self.load_count_substitutions()
        logger.info("Done with count base subsitutions (4/9)")
        self.load_indel_length()
        logger.info("Done with indel length (5/9)")
        self.load_annotation()
        logger.info("Done with effect annotations (6/9)")
        self.top_occurrences_loader.load()
        logger.info("Done with top occurrent variants (7/9)")
        self.ns_s_counts_loader.load()
        logger.info("Done with NS S counts (8/9)")
        self.variants_per_lineage_loader.load()
        logger.info("Done with variants per lineage (9/9)")

    def load_counts_variants_per_sample(self):

        # reads the data from the database
        database_rows_ena = self._read_count_variants_per_sample(data_source=DataSource.ENA)
        database_rows_gisaid = self._read_count_variants_per_sample(data_source=DataSource.GISAID)

        # delete all rows before starting
        self.session.query(PrecomputedVariantsPerSample).delete()
        self.session.commit()

        self.session.add_all(database_rows_ena + database_rows_gisaid)
        self.session.commit()
        logger.info("Added {} entries to {}".format(len(database_rows_ena) + len(database_rows_gisaid),
                                                    PrecomputedVariantsPerSample.__tablename__))

    def _read_count_variants_per_sample(self, data_source: DataSource):

        table_name = self.queries.get_variant_observation_klass(source=data_source.name).__tablename__

        sql_query = """
        select count(*) as count, counts.number_mutations, counts.variant_type, counts.gene_name from (
            select count(*) as number_mutations, variant_type, gene_name, sample from {table_name}
            group by variant_type, gene_name, sample)
            as counts group by counts.number_mutations, counts.variant_type, counts.gene_name;
            """.format(table_name=table_name)
        data = pd.read_sql_query(sql_query, self.session.bind)
        sql_query = """
        select count(*) as count, counts.number_mutations, counts.variant_type from (
            select count(*) as number_mutations, variant_type, sample from {table_name}
            group by variant_type, sample)
            as counts group by counts.number_mutations, counts.variant_type;
            """.format(table_name=table_name)
        data_without_gene = pd.read_sql_query(sql_query, self.session.bind)

        database_rows = []
        for index, row in data.iterrows():
            database_rows.append(PrecomputedVariantsPerSample(
                count=row["count"],
                number_mutations=row["number_mutations"],
                source=data_source.name,
                variant_type=row["variant_type"],
                gene_name=row["gene_name"] if row["gene_name"] is not None else "intergenic"
            ))
        for index, row in data_without_gene.iterrows():
            database_rows.append(PrecomputedVariantsPerSample(
                count=row["count"],
                number_mutations=row["number_mutations"],
                source=data_source.name,
                variant_type=row["variant_type"]
            ))
        return database_rows

    def load_count_substitutions(self):

        database_rows_ena = self._read_count_substitutions(data_source=DataSource.ENA)
        database_rows_gisaid = self._read_count_substitutions(data_source=DataSource.GISAID)

        # delete all rows before starting
        self.session.query(PrecomputedSubstitutionsCounts).delete()
        self.session.commit()

        self.session.add_all(database_rows_ena + database_rows_gisaid)
        self.session.commit()
        logger.info("Added {} entries to {}".format(len(database_rows_ena) + len(database_rows_gisaid),
                                                    PrecomputedSubstitutionsCounts.__tablename__))

    def _read_count_substitutions(self, data_source: DataSource):

        table_name = self.queries.get_variant_observation_klass(source=data_source.name).__tablename__

        sql_query = """
        select * from (select count(*) as count, reference, alternate, variant_type, gene_name 
            from {table_name}
            group by reference, alternate, variant_type, gene_name
            order by count desc) 
            as counts where counts.count > 10;
        """.format(table_name=table_name)
        data = pd.read_sql_query(sql_query, self.session.bind)
        sql_query = """
        select * from (select count(*) as count, reference, alternate, variant_type 
            from {table_name}
            group by reference, alternate, variant_type
            order by count desc) 
            as counts where counts.count > 10;
        """.format(table_name=table_name)
        data_without_gene = pd.read_sql_query(sql_query, self.session.bind)

        database_rows = []
        for index, row in data.iterrows():
            database_rows.append(PrecomputedSubstitutionsCounts(
                count=row["count"],
                reference=row["reference"],
                alternate=row["alternate"],
                source=data_source.name,
                variant_type=row["variant_type"],
                gene_name=row["gene_name"] if row["gene_name"] is not None else "intergenic"
            ))
        for index, row in data_without_gene.iterrows():
            database_rows.append(PrecomputedSubstitutionsCounts(
                count=row["count"],
                reference=row["reference"],
                alternate=row["alternate"],
                source=data_source.name,
                variant_type=row["variant_type"]
            ))
        return database_rows

    def load_indel_length(self):

        database_rows_ena = self._read_indel_length(data_source=DataSource.ENA)
        database_rows_gisaid = self._read_indel_length(data_source=DataSource.GISAID)

        # delete all rows before starting
        self.session.query(PrecomputedIndelLength).delete()
        self.session.commit()

        self.session.add_all(database_rows_ena + database_rows_gisaid)
        self.session.commit()
        logger.info("Added {} entries to {}".format(len(database_rows_ena) + len(database_rows_gisaid),
                                                    PrecomputedIndelLength.__tablename__))

    def _read_indel_length(self, data_source: DataSource):

        table_name = self.queries.get_variant_observation_klass(source=data_source.name).__tablename__

        sql_query = """
        select count(*) as count, length, gene_name 
            from {table_name}
            where variant_type != 'SNV'
            group by length, gene_name
            order by count asc;
        """.format(table_name=table_name)
        data = pd.read_sql_query(sql_query, self.session.bind)
        sql_query = """
        select count(*) as count, length, gene_name 
            from {table_name}
            where variant_type != 'SNV'
            group by length, gene_name
            order by count desc;
        """.format(table_name=table_name)
        data_without_gene = pd.read_sql_query(sql_query, self.session.bind)

        database_rows = []
        for index, row in data.iterrows():
            # add entries per gene
            database_rows.append(PrecomputedIndelLength(
                count=row["count"],
                length=row["length"],
                source=data_source.name,
                gene_name=row["gene_name"] if row["gene_name"] is not None else "intergenic"
            ))
        for index, row in data_without_gene.iterrows():
            # add entries per gene
            database_rows.append(PrecomputedIndelLength(
                count=row["count"],
                length=row["length"],
                source=data_source.name
            ))
        return database_rows

    def load_annotation(self):

        database_rows_ena = self._read_annotations(data_source=DataSource.ENA)
        database_rows_gisaid = self._read_annotations(data_source=DataSource.GISAID)

        # delete all rows before starting
        self.session.query(PrecomputedAnnotation).delete()
        self.session.commit()

        self.session.add_all(database_rows_ena + database_rows_gisaid)
        self.session.commit()
        logger.info("Added {} entries to {}".format(len(database_rows_ena) + len(database_rows_gisaid), PrecomputedAnnotation.__tablename__))

    def _read_annotations(self, data_source: DataSource):

        table_name = self.queries.get_variant_observation_klass(source=data_source.name).__tablename__

        sql_query = """
        select count(*) as count, annotation_highest_impact, gene_name 
            from {table_name}
            where annotation_highest_impact is not null
            group by annotation_highest_impact, gene_name
            order by count asc;
        """.format(table_name=table_name)
        data = pd.read_sql_query(sql_query, self.session.bind)
        sql_query = """
        select count(*) as count, annotation_highest_impact 
            from {table_name}
            where annotation_highest_impact is not null
            group by annotation_highest_impact
            order by count asc;
        """.format(table_name=table_name)
        data_without_gene = pd.read_sql_query(sql_query, self.session.bind)
        # stores the precomputed data
        database_rows = []
        for index, row in data.iterrows():
            # add entries per gene
            database_rows.append(PrecomputedAnnotation(
                count=row["count"],
                annotation=row["annotation_highest_impact"],
                source=data_source.name,
                gene_name=row["gene_name"] if row["gene_name"] is not None else "intergenic"
            ))
        for index, row in data_without_gene.iterrows():
            # add entries per gene
            database_rows.append(PrecomputedAnnotation(
                count=row["count"],
                annotation=row["annotation_highest_impact"],
                source=data_source.name
            ))
        return database_rows

    def load_table_counts(self):

        count_variants_ena = self.queries.count_variants(cache=False, source=DataSource.ENA.name)
        count_variants_gisaid = self.queries.count_variants(cache=False, source=DataSource.GISAID.name)
        count_samples_ena = self.queries.count_samples(source=DataSource.ENA.name, cache=False)
        count_samples_gisaid = self.queries.count_samples(source=DataSource.GISAID.name, cache=False)
        count_variant_observations_ena = self.queries.count_variant_observations(
            source=DataSource.ENA.name, cache=False)
        count_variant_observations_gisaid = self.queries.count_variant_observations(
            source=DataSource.GISAID.name, cache=False)
        count_subclonal_variant_observations = self.queries.count_subclonal_variant_observations(cache=False)
        count_subclonal_variant_unique = self.queries.count_unique_subclonal_variant(cache=False)
        count_subclonal_variant_unique_only_subclonal = self.queries.count_unique_only_subclonal_variant(cache=False)
        count_countries_ena = self.queries.count_countries(source=DataSource.ENA.name, cache=False)
        count_countries_gisaid = self.queries.count_countries(source=DataSource.GISAID.name, cache=False)

        # delete all rows before starting
        self.session.query(PrecomputedTableCounts).delete()
        self.session.commit()

        database_rows = [
            PrecomputedTableCounts(
                table=Variant.__name__, count=count_variants_ena,
                factor=PrecomputedTableCounts.FACTOR_SOURCE, value=DataSource.ENA.name),
            PrecomputedTableCounts(
                table=GisaidVariant.__name__, count=count_variants_gisaid,
                factor=PrecomputedTableCounts.FACTOR_SOURCE, value=DataSource.GISAID.name),
            PrecomputedTableCounts(
                table=VariantObservation.__name__, count=count_variant_observations_ena,
                factor=PrecomputedTableCounts.FACTOR_SOURCE, value=DataSource.ENA.name),
            PrecomputedTableCounts(
                table=GisaidVariantObservation.__name__, count=count_variant_observations_gisaid,
                factor=PrecomputedTableCounts.FACTOR_SOURCE, value=DataSource.GISAID.name),
            PrecomputedTableCounts(
                table=SubclonalVariantObservation.__name__, count=count_subclonal_variant_observations),
            PrecomputedTableCounts(
                table=SubclonalVariantObservation.__name__ + "_unique", count=count_subclonal_variant_unique),
            PrecomputedTableCounts(
                table=SubclonalVariantObservation.__name__ + "_unique_only_subclonal",
                count=count_subclonal_variant_unique_only_subclonal),
            PrecomputedTableCounts(
                table=SampleEna.__name__, count=count_samples_ena,
                factor=PrecomputedTableCounts.FACTOR_SOURCE, value=DataSource.ENA.name),
            PrecomputedTableCounts(
                table=SampleGisaid.__name__, count=count_samples_gisaid,
                factor=PrecomputedTableCounts.FACTOR_SOURCE, value=DataSource.GISAID.name),
            PrecomputedTableCounts(
                table=PrecomputedTableCounts.VIRTUAL_TABLE_COUNTRY, count=count_countries_ena,
                factor=PrecomputedTableCounts.FACTOR_SOURCE, value=DataSource.ENA.name),
            PrecomputedTableCounts(
                table=PrecomputedTableCounts.VIRTUAL_TABLE_COUNTRY, count=count_countries_gisaid,
                factor=PrecomputedTableCounts.FACTOR_SOURCE, value=DataSource.GISAID.name),
        ]

        if len(database_rows) > 0:
            self.session.add_all(database_rows)
            self.session.commit()
        logger.info("Added {} entries to {}".format(len(database_rows), PrecomputedTableCounts.__tablename__))

    def load_variant_abundance_histogram(self):

        database_rows = self._read_variant_abundance_histogram()

        # delete all rows before starting
        self.session.query(PrecomputedVariantAbundanceHistogram).delete()
        self.session.commit()

        if len(database_rows) > 0:
            self.session.add_all(database_rows)
            self.session.commit()
        logger.info("Added {} entries to {}".format(
            len(database_rows), PrecomputedVariantAbundanceHistogram.__tablename__))

    def _read_variant_abundance_histogram(self):
        histograms = []
        for bin_size in BIN_SIZE_VALUES:
            histogram = self.queries.get_variant_abundance_histogram(
                bin_size=bin_size, source=DataSource.ENA.name, cache=False)
            if histogram is not None:
                histogram["bin_size"] = bin_size
                histogram["source"] = DataSource.ENA
                histograms.append(histogram)
        for bin_size in BIN_SIZE_VALUES:
            histogram = self.queries.get_variant_abundance_histogram(
                bin_size=bin_size, source=DataSource.GISAID.name, cache=False)
            if histogram is not None:
                histogram["bin_size"] = bin_size
                histogram["source"] = DataSource.GISAID
                histograms.append(histogram)
        database_rows = []
        if len(histograms) > 0:
            for index, row in pd.concat(histograms).iterrows():
                # add entries per gene
                database_rows.append(PrecomputedVariantAbundanceHistogram(
                    position_bin=row["position_bin"],
                    count_unique_variants=row["count_unique_variants"],
                    count_variant_observations=row["count_variant_observations"],
                    bin_size=row["bin_size"],
                    source=row["source"],
                ))
        return database_rows



import os

import sqlalchemy
from sqlalchemy.orm import Session
from covigator.database.model import Conservation
from logzero import logger
import pandas as pd


class ConservationLoader:

    CONSERVATION_FILENAME = "wuhCor1.mutDepletionConsHMM.bed"
    CONSERVATION_SARBECOVIRUS_FILENAME = "wuhCor1.mutDepletionSarbecovirusConsHMM.bed"
    CONSERVATION_VERTEBRATES_FILENAME = "wuhCor1.mutDepletionVertebrateCoVConsHMM.bed"

    def __init__(self, session: Session):

        # these files were downloaded from https://github.com/ernstlab/ConsHMM_CoV
        self.conservation_filename = os.path.join(
            os.path.abspath(os.path.dirname(__file__)), self.CONSERVATION_FILENAME)
        self.conservation_sarbecovirus_filename = os.path.join(
            os.path.abspath(os.path.dirname(__file__)), self.CONSERVATION_SARBECOVIRUS_FILENAME)
        self.conservation_vertebrates_filename = os.path.join(
            os.path.abspath(os.path.dirname(__file__)), self.CONSERVATION_VERTEBRATES_FILENAME)

        self.session = session

    def load_data(self):
        # read conservation data from BED files
        conservation = pd.read_csv(
            self.conservation_filename, skiprows=1, names=["chromosome", "start", "end", "conservation"], sep="\t")
        conservation_sarbecovirus = pd.read_csv(
            self.conservation_sarbecovirus_filename, skiprows=1,
            names=["chromosome", "start", "end", "conservation_sarbecovirus"], sep="\t")
        conservation_vertebrates = pd.read_csv(
            self.conservation_vertebrates_filename, skiprows=1,
            names=["chromosome", "start", "end", "conservation_vertebrates"], sep="\t")

        # joins conservation data in a single table
        data = conservation.set_index(["chromosome", "start", "end"]).join(
            conservation_sarbecovirus.set_index(["chromosome", "start", "end"])
        ).join(
            conservation_vertebrates.set_index(["chromosome", "start", "end"])
        )
        data.reset_index(inplace=True)
        data.fillna(0, inplace=True)

        # persist the conservation table
        metadata = sqlalchemy.schema.MetaData(bind=self.session.bind, reflect=True)
        conservation_table = sqlalchemy.Table(Conservation.__tablename__, metadata, autoload=True)
        self.session.execute(conservation_table.insert(), data.to_dict(orient='records'))
        self.session.commit()

        logger.info("Loaded into the database {} conservation data points".format(data.shape[0]))



from itertools import combinations
from typing import Union

from sqlalchemy.orm import Session
from covigator.database.model import VariantCooccurrence, DataSource
from logzero import logger
from covigator.database.queries import Queries
from sqlalchemy.exc import IntegrityError


class CooccurrenceMatrixException(Exception):
    pass


class CooccurrenceMatrix:

    def compute(self, run_accession: str, source: DataSource, session: Session):

        assert run_accession is not None or run_accession == "", "Missing sample identifier"
        assert session is not None, "Missing DB session"

        queries = Queries(session=session)
        sample_id = run_accession
        logger.info("Processing cooccurrent variants for sample {}".format(sample_id))

        # the order by position is important to ensure we store only half the matrix and the same half of the matrix
        variants = queries.get_variants_by_sample(sample_id, source=source.name)
        failed_variants = []

        # process all pairwise combinations without repetitions including the diagoonal
        for (variant_one, variant_two) in list(combinations(variants, 2)) + list(zip(variants, variants)):
            try:
                variant_cooccurrence = queries.get_variant_cooccurrence(variant_one, variant_two)
                if variant_cooccurrence is None:
                    variant_cooccurrence = VariantCooccurrence(
                        variant_id_one=variant_one.variant_id,
                        variant_id_two=variant_two.variant_id,
                        count=1
                    )
                    session.add(variant_cooccurrence)
                    session.commit()
                else:
                    # NOTE: it is important to increase the counter like this to avoid race conditions
                    # the increase happens in the database server and not in python
                    # see https://stackoverflow.com/questions/2334824/how-to-increase-a-counter-in-sqlalchemy
                    variant_cooccurrence.count = VariantCooccurrence.count + 1
            except IntegrityError:
                session.rollback()
                failed_variants.append((variant_one, variant_two))

        # tries again the failed variants as these are expected to be there now
        for (variant_one, variant_two) in failed_variants:
            variant_cooccurrence = queries.get_variant_cooccurrence(variant_one, variant_two)
            if variant_cooccurrence is None:
                raise CooccurrenceMatrixException("Some cooccurrent variants failed to be persisted twice")
            variant_cooccurrence.count = VariantCooccurrence.count + 1



from contextlib import contextmanager
import tenacity
from logzero import logger
from sqlalchemy import create_engine, text
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker, Session
from tenacity import wait_exponential, stop_after_attempt
from covigator.configuration import Configuration
from covigator.database.model import Base, Gene, Conservation
from covigator.exceptions import CovigatorDatabaseConnectionException
from covigator.references.conservation import ConservationLoader
from covigator.references.gene_annotations import GeneAnnotationsLoader


class Database:

    def __init__(self, config: Configuration = None, test=False, verbose=False, initialize=False):
        if test:
            # connects to the test postgres in the CI environment
            #db_uri = "postgresql+psycopg2://%s:%s@%s/%s" % ("test_user", "test_pass", "postgres", "test_db")
            #self.engine: Engine = create_engine(db_uri, echo=verbose)
            initialize = True   # does not make sense not to initialize the DB in the test environment

        db_uri = "postgresql+psycopg2://%s:%s@%s:%s/%s" % (config.db_user, config.db_password, config.db_host,
                                                           config.db_port, config.db_name)
        # these are the default SQLAlchemy values, this values may need to be increased for the dashboard
        self.engine: Engine = create_engine(db_uri, pool_size=config.db_pool_size,
                                            max_overflow=config.db_max_overflow, echo=verbose)
        self.engine.connect()
        self.Session = sessionmaker(bind=self.engine, autoflush=False)
        # avoids initialisation if we know it has already been done
        if initialize:
            self.create_database()

    def create_database(self):
        # this creates all tables in the database (when it exists nothing happens)
        Base.metadata.create_all(self.engine)
        self.initialise_database()
        logger.info("Database initialized")

    def initialise_database(self):
        session = self.get_database_session()
        # loads reference genome if not set
        if session.query(Gene).count() == 0:
            GeneAnnotationsLoader(session).load_data()
        if session.query(Conservation).count() == 0:
            ConservationLoader(session).load_data()

    def get_database_session(self) -> Session:
        return self.Session()


@contextmanager
def session_scope(config: Configuration = None, database: Database = None, initialize=False, test=False) -> Session:
    """Provide a transactional scope around a series of operations.
    Either config or database must be provided!
    """
    if database is None and config is None:
        raise CovigatorDatabaseConnectionException(
            "Must provide either a Configuration object or a database connection to obtain a session")
    if database is None:
        database = Database(config=config, initialize=initialize, test=test)
    session = database.get_database_session()
    try:
        yield session
        session.commit()
    except BaseException as e:
        session.rollback()
        raise e
    finally:
        session.close()
        database.engine.dispose()


@tenacity.retry(wait=wait_exponential(multiplier=2, min=1, max=10), stop=stop_after_attempt(5))
def get_database(config: Configuration, initialize=False, verbose=False) -> Database:
    try:
        database = Database(config=config, initialize=initialize, verbose=verbose)
        session = database.get_database_session()
        stmt = text("SELECT 1")
        session.execute(stmt)
        logger.info("Database connected!")
    except Exception as e:
        logger.error("Connection to database failed, retrying...")
        raise CovigatorDatabaseConnectionException(e)
    return database



from dataclasses import dataclass

import pycountry
import pycountry_convert


@dataclass
class ParsedCountry:
    country_raw: str
    country: str
    country_alpha_2: str
    country_alpha_3: str
    continent_alpha_2: str
    continent: str


empty_country = ParsedCountry(
    country_raw="",
    country="Not available",
    country_alpha_2="None",    # don't use NA as that is the id of Namibia
    country_alpha_3="None",
    continent_alpha_2="None",
    continent="None"
)


class CountryParser:

    def __init__(self):
        self.cache = {}

    def parse_country(self, country_raw: str) -> ParsedCountry:
        try:
            if country_raw is None or country_raw.strip() == "":
                parsed_country = empty_country
            else:
                parsed_country = self.cache.get(country_raw)
                if parsed_country is None:
                    match = pycountry.countries.search_fuzzy(country_raw)[0]
                    continent = pycountry_convert.country_alpha2_to_continent_code(match.alpha_2)
                    parsed_country = ParsedCountry(
                        country_raw=country_raw,
                        country=match.name,
                        country_alpha_2=match.alpha_2,
                        country_alpha_3=match.alpha_3,
                        continent_alpha_2=continent,
                        continent=pycountry_convert.convert_continent_code_to_continent_name(continent))
                    self.cache[country_raw] = parsed_country
        except (LookupError, AttributeError):
            parsed_country = empty_country
            parsed_country.country_raw = country_raw
        return parsed_country



import pkg_resources
import covigator.tests
from covigator.database.model import Variant, VariantObservation, DataSource, SubclonalVariantObservation, \
    SampleGisaid, SampleEna, GisaidVariant, GisaidVariantObservation, LowFrequencyVariantObservation, SubclonalVariant, \
    LowFrequencyVariant
from covigator.pipeline.vcf_loader import VcfLoader
from covigator.tests.unit_tests.abstract_test import AbstractTest


class VcfLoaderTests(AbstractTest):

    def setUp(self) -> None:
        self.sample_ena = SampleEna(run_accession="TEST1", fastq_ftp="something", fastq_md5="else", num_fastqs=2)
        self.sample_gisaid = SampleGisaid(run_accession="TEST2")
        self.session.add_all([self.sample_ena, self.sample_gisaid])
        self.session.commit()

    def test_vcf_loader(self):
        vcf_file = pkg_resources.resource_filename(covigator.tests.__name__, "resources/snpeff.vcf")
        VcfLoader().load(vcf_file, run_accession=self.sample_ena.run_accession, source=DataSource.ENA, session=self.session)
        self.session.commit()
        self.assertEqual(self.session.query(Variant).count(), 1)
        self.assertEqual(self.session.query(VariantObservation).count(), 1)
        self.assertEqual(self.session.query(SubclonalVariant).count(), 1)
        self.assertEqual(self.session.query(SubclonalVariantObservation).count(), 1)
        self.assertEqual(self.session.query(LowFrequencyVariant).count(), 1)
        self.assertEqual(self.session.query(LowFrequencyVariantObservation).count(), 1)
        variant = self.session.query(Variant).first()
        self.assertEqual(variant.chromosome, "MN908947.3")
        self.assertEqual(variant.position, 23403)
        self.assertEqual(variant.reference, "A")
        self.assertEqual(variant.alternate, "G")
        self.assertEqual(variant.gene_name, "S")
        variant_observation = self.session.query(VariantObservation).first()
        self.assertEqual(variant_observation.sample, self.sample_ena.run_accession)
        self.assertEqual(variant_observation.chromosome, "MN908947.3")
        self.assertEqual(variant_observation.position, 23403)
        self.assertEqual(variant_observation.reference, "A")
        self.assertEqual(variant_observation.alternate, "G")
        self.assertIsNone(variant_observation.dp)
        self.assertIsNone(variant_observation.vaf)

    def test_vcf_loader_without_dp4(self):
        vcf_file = pkg_resources.resource_filename(covigator.tests.__name__, "resources/snpeff_without_dp4.vcf")
        VcfLoader().load(vcf_file, run_accession=self.sample_ena.run_accession, source=DataSource.ENA, session=self.session)
        self.session.commit()
        self.assertEqual(self.session.query(Variant).count(), 1)
        self.assertEqual(self.session.query(VariantObservation).count(), 1)

    def test_vcf_loader_without_annotation(self):
        vcf_file = pkg_resources.resource_filename(covigator.tests.__name__, "resources/snpeff_without_annotation.vcf")
        VcfLoader().load(vcf_file, run_accession=self.sample_ena.run_accession, source=DataSource.ENA, session=self.session)
        self.session.commit()
        self.assertEqual(self.session.query(Variant).count(), 1)
        self.assertEqual(self.session.query(VariantObservation).count(), 1)

    def test_vcf_loader_with_insertion(self):
        vcf_file = pkg_resources.resource_filename(covigator.tests.__name__, "resources/snpeff_with_insertion.vcf")
        VcfLoader().load(vcf_file, run_accession=self.sample_ena.run_accession, source=DataSource.ENA, session=self.session)
        self.session.commit()
        self.assertEqual(self.session.query(Variant).count(), 1)
        self.assertEqual(self.session.query(VariantObservation).count(), 1)

    def test_vcf_loader_with_deletion(self):
        vcf_file = pkg_resources.resource_filename(covigator.tests.__name__, "resources/snpeff_with_deletion.vcf")
        VcfLoader().load(vcf_file, run_accession=self.sample_ena.run_accession, source=DataSource.ENA, session=self.session)
        self.session.commit()
        self.assertEqual(self.session.query(Variant).count(), 1)
        self.assertEqual(self.session.query(VariantObservation).count(), 1)

    def test_vcf_loader_gisaid(self):
        vcf_file = pkg_resources.resource_filename(covigator.tests.__name__, "resources/snpeff.vcf")
        VcfLoader().load(vcf_file, run_accession=self.sample_gisaid.run_accession, source=DataSource.GISAID, session=self.session)
        self.session.commit()
        self.assertEqual(self.session.query(GisaidVariant).count(), 3)
        self.assertEqual(self.session.query(GisaidVariantObservation).count(), 3)
        variant = self.session.query(GisaidVariant).first()
        self.assertEqual(variant.chromosome, "MN908947.3")
        self.assertEqual(variant.position, 23403)
        self.assertEqual(variant.reference, "A")
        self.assertEqual(variant.alternate, "G")
        self.assertEqual(variant.gene_name, "S")
        variant_observation = self.session.query(GisaidVariantObservation).first()
        self.assertEqual(variant_observation.sample, self.sample_gisaid.run_accession)
        self.assertEqual(variant_observation.chromosome, "MN908947.3")
        self.assertEqual(variant_observation.position, 23403)
        self.assertEqual(variant_observation.reference, "A")
        self.assertEqual(variant_observation.alternate, "G")

    def test_vcf_loader_vafator(self):
        vcf_file = pkg_resources.resource_filename(covigator.tests.__name__, "resources/test.lofreq.vcf.gz")
        VcfLoader().load(vcf_file, run_accession=self.sample_ena.run_accession, source=DataSource.ENA, session=self.session)
        self.session.commit()
        self.assertEqual(self.session.query(Variant).count(), 2)
        self.assertEqual(self.session.query(VariantObservation).count(), 2)
        self.assertEqual(self.session.query(SubclonalVariant).count(), 1)
        self.assertEqual(self.session.query(SubclonalVariantObservation).count(), 1)
        self.assertEqual(self.session.query(LowFrequencyVariant).count(), 8)
        self.assertEqual(self.session.query(LowFrequencyVariantObservation).count(), 8)

        # NOTE: this test VCF was computed with thresholds of 0.2 and 0.8
        for vo in self.session.query(VariantObservation).all():
            self.assertGreaterEqual(vo.vaf, 0.8)
            self.assertGreater(vo.dp, 0)
            self.assertGreater(vo.ac, 0)

        for vo in self.session.query(SubclonalVariantObservation).all():
            self.assertGreaterEqual(vo.vaf, 0.2)
            self.assertLess(vo.vaf, 0.8)
            self.assertGreater(vo.dp, 0)
            self.assertGreater(vo.ac, 0)

        for vo in self.session.query(LowFrequencyVariantObservation).all():
            self.assertLess(vo.vaf, 0.2)
            self.assertGreater(vo.dp, 0)
            self.assertGreater(vo.ac, 0)

    def test_vcf_loader_gisaid_2(self):
        vcf_file = pkg_resources.resource_filename(covigator.tests.__name__, "resources/test.assembly.vcf.gz")
        VcfLoader().load(vcf_file, run_accession=self.sample_gisaid.run_accession, source=DataSource.GISAID, session=self.session)
        self.session.commit()
        self.assertEqual(self.session.query(GisaidVariant).count(), 13)
        self.assertEqual(self.session.query(GisaidVariantObservation).count(), 13)

        # NOTE: this test VCF was computed with thresholds of 0.2 and 0.8
        for vo in self.session.query(GisaidVariantObservation).all():
            self.assertIsNone(vo.vaf)
            self.assertIsNone(vo.dp)
            self.assertIsNone(vo.ac)



from unittest import TestCase

from covigator.configuration import Configuration
from covigator.references.gene_annotations import GeneAnnotationsLoader


class GeneAnnotationsLoaderTest(TestCase):

    def setUp(self) -> None:
        self.loader = GeneAnnotationsLoader()

    def test_load(self):
        self.loader.load_data()



import dash_core_components as dcc
import dash_bootstrap_components as dbc
import dash_html_components as html
import plotly

from covigator.dashboard.figures.figures import PLOTLY_CONFIG, MARGIN, TEMPLATE
from covigator.dashboard.tabs import get_mini_container, print_number, print_date
from covigator.database.model import DataSource, SAMPLE_GISAID_TABLE_NAME, JobStatus
from covigator.database.queries import Queries
import pandas as pd
import plotly.express as px


def get_tab_dataset_gisaid(queries: Queries):
    count_samples = queries.count_samples(source=DataSource.GISAID.name)

    return dbc.CardBody(
        children=[
            get_gisaid_overview_tab_left_bar(queries, count_samples),
            html.Div(
                className="one column",
                children=[html.Br()]),
            get_gisaid_overview_tab_graphs(queries=queries, count_samples=count_samples)
        ])


def get_gisaid_overview_tab_graphs(queries, count_samples):
    return html.Div(
        className="nine columns",
        children=get_dataset_gisaid_tab_graphs(queries, count_samples=count_samples))


def get_gisaid_overview_tab_left_bar(queries: Queries, count_samples):
    count_variants = queries.count_variants(source=DataSource.GISAID.name)
    count_variant_observations = queries.count_variant_observations(source=DataSource.GISAID.name)
    date_of_first_gisaid_sample = queries.get_date_of_first_sample(source=DataSource.GISAID)
    date_of_most_recent_gisaid_sample = queries.get_date_of_most_recent_sample(source=DataSource.GISAID)

    return html.Div(
        className="two columns",
        children=[
            html.Br(),
            dcc.Markdown("""
                The GISAID database (https://www.gisaid.org/) provides DNA assemblies, geographical information and
                other metadata about SARS-CoV-2 samples. 
                The processing pipeline runs alignment to the reference genome (bioypthon), 
                variant calling (custom code), normalization (vt and BCFtools), annotation (SnpEff) 
                and finally lineage determination (pangolin). 
                """, style={"font-size": 16}),
            html.Br(),
            html.Div(
                html.Span(
                    children=[
                        get_mini_container(
                            title="Samples",
                            value=print_number(count_samples)),
                        html.Br(),
                        html.Br(),
                        get_mini_container(
                            title="First sample",
                            value=print_date(date_of_first_gisaid_sample)),
                        html.Br(),
                        html.Br(),
                        get_mini_container(
                            title="Latest sample",
                            value=print_date(date_of_most_recent_gisaid_sample)),
                        html.Br(),
                        html.Br(),
                        get_mini_container(
                            title="Unique mutations",
                            value=print_number(count_variants)),
                        html.Br(),
                        html.Br(),
                        get_mini_container(
                            title="Mutation calls",
                            value=print_number(count_variant_observations)),
                    ])
            ),
            html.Br(),
            dcc.Markdown("""
                    There are two sample exclusion criteria: 
                    * Horizontal coverage below 20 % of the reference genome
                    * A ratio of ambiguous bases greater than 0.2 
                    
                    Furthermore, all mutations involving ambiguous bases are excluded. 
                    """, style={"font-size": 16}),
        ])


def get_dataset_gisaid_tab_graphs(queries: Queries, count_samples):

    return html.Div(
        children=[
            html.Br(),
            html.Div(children=[
                html.Div(children=get_plot_coverage(queries),
                         className="six columns", style={"margin-left": 0, "margin-right": "1%", "width": "48%"}),
                html.Div(children=get_plot_bad_bases_ratio(queries, count_samples),
                         className="six columns", style={"margin-left": 0, "margin-right": "1%", "width": "48%"}),
            ]),
        ])


def get_plot_coverage(queries: Queries):
    sql_query = """
    select count(*), (sequence_length::float / 29903 * 100)::int as coverage
    from {table} 
    where status = '{status}'
    group by coverage
    """.format(table=SAMPLE_GISAID_TABLE_NAME, status=JobStatus.FINISHED.name)
    data = pd.read_sql_query(sql_query, queries.session.bind)
    fig = px.bar(data_frame=data, x="coverage", y="count", log_y=True, color="count",
                 color_continuous_scale=plotly.colors.sequential.Brwnyl)
    fig.update_layout(
        margin=MARGIN,
        template=TEMPLATE,
        yaxis={'title': "Num. of samples (log)"},
        xaxis={'title': "Horizontal coverage (%)"},
        legend={'title': None}
    )
    return [
        dcc.Graph(figure=fig, config=PLOTLY_CONFIG),
        dcc.Markdown("""
        **Horizontal coverage (%)**
        
        Horizontal coverage is estimated as the sequence length divided by the reference genome length.
        Some samples have thus an horizontal coverage larger than 100 %. 
        """)
    ]


def get_plot_bad_bases_ratio(queries: Queries, count_samples):
    sql_query = """
    select count(*), ((count_n_bases + count_ambiguous_bases)::float / sequence_length * 100)::int as bad_bases_ratio
    from {table} 
    where status = '{status}'
    group by bad_bases_ratio
    """.format(table=SAMPLE_GISAID_TABLE_NAME, status=JobStatus.FINISHED.name)
    data = pd.read_sql_query(sql_query, queries.session.bind)
    fig = px.bar(data_frame=data, x="bad_bases_ratio", y="count", log_y=True, color="count",
                 color_continuous_scale=plotly.colors.sequential.Brwnyl)
    fig.update_layout(
        margin=MARGIN,
        template=TEMPLATE,
        yaxis={'title': "Num. of samples (log)"},
        xaxis={'title': "Ratio of N and ambiguous bases (%)"},
        legend={'title': None}
    )


    return [
        dcc.Graph(figure=fig, config=PLOTLY_CONFIG),
        dcc.Markdown("""
        **Ratio of N and ambiguous bases (%)**

        The ratio of N and ambiguous bases over the whole sequence length measures the quality of the assembly sequence.
        {} % of samples have a ratio <= 5 %.
        """.format(round(float(data[data["bad_bases_ratio"] <= 5]["count"].sum()) / count_samples), 1))
    ]


from typing import List

import pandas as pd
from logzero import logger
from sqlalchemy.orm import Session
from covigator import SYNONYMOUS_VARIANT, MISSENSE_VARIANT
from covigator.database.model import DataSource, PrecomputedSynonymousNonSynonymousCounts, RegionType
from covigator.database.queries import Queries


class NsSCountsLoader:

    def __init__(self, session: Session):
        self.session = session
        self.queries = Queries(session=self.session)

    def load(self):

        counts_by_gene = self._load_counts(region=RegionType.GENE)
        counts_by_domain = self._load_counts(region=RegionType.DOMAIN)
        database_rows = counts_by_gene + counts_by_domain

        # delete all rows before starting
        self.session.query(PrecomputedSynonymousNonSynonymousCounts).delete()
        self.session.commit()

        if len(database_rows) > 0:
            self.session.add_all(database_rows)
            self.session.commit()
        logger.info("Added {} entries to {}".format(len(database_rows),
                                                    PrecomputedSynonymousNonSynonymousCounts.__tablename__))

    def _load_counts(self, region: RegionType) -> List[PrecomputedSynonymousNonSynonymousCounts]:
        data_s_ena = self._count_variant_observations_by_source_annotation_and_region(
            source=DataSource.ENA, annotation=SYNONYMOUS_VARIANT, region=region)
        data_ns_ena = self._count_variant_observations_by_source_annotation_and_region(
            source=DataSource.ENA, annotation=MISSENSE_VARIANT, region=region)
        data_s_gisaid = self._count_variant_observations_by_source_annotation_and_region(
            source=DataSource.GISAID, annotation=SYNONYMOUS_VARIANT, region=region)
        data_ns_gisaid = self._count_variant_observations_by_source_annotation_and_region(
            source=DataSource.GISAID, annotation=MISSENSE_VARIANT, region=region)

        data_ena = pd.merge(
            left=data_s_ena, right=data_ns_ena, on=["month", "region_name", "country"], how='outer').fillna(0)
        data_gisaid = pd.merge(
            left=data_s_gisaid, right=data_ns_gisaid, on=["month", "region_name", "country"], how='outer').fillna(0)

        database_rows = []
        if data_ena is not None:
            database_rows.extend(self._dataframe_to_model(
                data=data_ena, source=DataSource.ENA, region=region))
        if data_gisaid is not None:
            database_rows.extend(self._dataframe_to_model(
                data=data_gisaid, source=DataSource.GISAID, region=region))

        return database_rows

    def _dataframe_to_model(self, data, source: DataSource, region: RegionType):

        database_rows = []
        for index, row in data.iterrows():
            database_rows.append(self._row_to_model(row, source, region))

        if region == RegionType.GENE:
            # if we are processing genes then we calculate also for the coding region overall
            coding_region_ns = {}
            coding_region_s = {}
            for index, row in data.iterrows():
                month = row['month']
                country = row['country']
                ns = row["ns"]
                s = row["s"]
                coding_region_ns[(month, country)] = coding_region_ns.get((month, country), 0) + ns
                coding_region_s[(month, country)] = coding_region_s.get((month, country), 0) + s
            for month, country in coding_region_ns.keys():
                database_rows.append(PrecomputedSynonymousNonSynonymousCounts(
                    month=month,
                    region_type=RegionType.CODING_REGION,
                    country=country,
                    ns=coding_region_ns.get((month, country), 0),
                    s=coding_region_s.get((month, country), 0),
                    source=source
                ))

        return database_rows

    def _row_to_model(self, row, source, region: RegionType):
        # add entries per gene
        return PrecomputedSynonymousNonSynonymousCounts(
            month=row['month'],
            region_type=region,
            region_name=row["region_name"],
            country=row['country'],
            ns=row["ns"],
            s=row["s"],
            source=source
        )

    def _count_variant_observations_by_source_annotation_and_region(
            self, source: DataSource, annotation: str, region: RegionType):

        variant_observation_table = self.queries.get_variant_observation_klass(source=source.name).__tablename__
        sample_table = self.queries.get_sample_klass(source=source.name).__tablename__

        sql_query = """
                select count(*) as {count_name}, date_trunc('month', s.collection_date) as month, 
                    vo.{region_field} as region_name, s.country 
                from {variant_observation_table} as vo join {sample_table} as s on vo.sample = s.run_accession 
                where vo.annotation_highest_impact = '{annotation}'
                    and s.collection_date is not null and vo.{region_field} is not null
                group by date_trunc('month', s.collection_date), vo.{region_field}, s.country;
                """.format(variant_observation_table=variant_observation_table,
                           sample_table=sample_table,
                           region_field="gene_name" if region == RegionType.GENE else "pfam_name",
                           count_name="s" if annotation == SYNONYMOUS_VARIANT else "ns",
                           annotation=annotation)
        data = pd.read_sql_query(sql_query, self.session.bind)
        return data



from typing import List

import colorlover
import dash_table
import pandas as pd
from logzero import logger
from plotly.subplots import make_subplots
from sqlalchemy import and_

from covigator.dashboard.figures.figures import Figures, PLOTLY_CONFIG, MARGIN, TEMPLATE, STYLES_STRIPPED, STYLE_HEADER, \
    STYLE_CELL
import plotly.express as px
import dash_core_components as dcc

from covigator.database.model import PrecomputedVariantsPerLineage, Variant


class LineageFigures(Figures):

    def get_lineages_plot(self, data_source: str, countries=None, lineages=None):
        logger.debug("Getting data on samples by country...")
        data = self.queries.get_accumulated_lineages_by_country(
            data_source=data_source, countries=countries, lineages=lineages)
        graph = dcc.Markdown("""**No data for the current selection**""")
        if data is not None and data.shape[0] > 0:
            logger.debug("Prepare plot on samples by lineage...")
            lineages = list(data.sort_values("cumsum", ascending=False).lineage.unique())

            fig = make_subplots(rows=2, cols=1)

            fig1 = px.area(
                data, x="date", y="cumsum", color="lineage",
                category_orders={"lineage": lineages[::-1]},
                labels={"ratio_per_date": "% daily samples", "cumsum": "num. samples", "count": "increment"},
                hover_data=["ratio_per_date", "count"],
                color_discrete_sequence=px.colors.qualitative.Vivid)
            fig1.update_traces(line=dict(width=0.5), showlegend=False)

            fig2 = px.area(
                data, x="date", y="ratio_per_date", color="lineage",
                category_orders={"lineage": lineages[::-1]},
                labels={"ratio_per_date": "% daily samples", "cumsum": "num. samples", "count": "increment"},
                hover_data=["cumsum", "count"],
                color_discrete_sequence=px.colors.qualitative.Vivid)
            fig2.update_traces(line=dict(width=0.5))

            for trace in fig1["data"]:
                fig.append_trace(trace, row=1, col=1)
            for trace in fig2["data"]:
                fig.append_trace(trace, row=2, col=1)

            fig.update_layout(
                margin=MARGIN,
                template=TEMPLATE,
                legend={'traceorder': 'reversed', 'title': None},
                xaxis={'title': None},
                yaxis={
                    'title': 'num. samples'
                },
                yaxis2={
                    'tickformat': ',.0%',
                    'range': [0, 1],
                    'title': '% samples'
                },
                height=700,
            )
            fig.update_xaxes(showspikes=True)

            top_lineages = lineages[0: min(5, len(lineages))]
            top_lineages_and_cumsum = data[data.lineage.isin(top_lineages)][["lineage", "cumsum"]] \
                .groupby("lineage").max().reset_index().sort_values("cumsum", ascending=False)
            top_lineages_tooltip = list(
                top_lineages_and_cumsum.apply(lambda x: "{} ({})".format(x[0], int(x[1])), axis=1))

            graph = [
                dcc.Graph(figure=fig, config=PLOTLY_CONFIG),
                dcc.Markdown("""
                **Samples by lineages**
                
                Top {} lineages: {}.
                """.format(len(top_lineages_tooltip),
                           ", ".join(top_lineages_tooltip)))
            ]
        return graph

    def discrete_background_color_bins(self, df, n_bins=5, columns='all', colors='Blues'):
        bounds = [i * (1.0 / n_bins) for i in range(n_bins + 1)]
        if columns == 'all':
            if 'id' in df:
                df_numeric_columns = df.select_dtypes('number').drop(['id'], axis=1)
            else:
                df_numeric_columns = df.select_dtypes('number')
        else:
            df_numeric_columns = df[columns]
        df_max = df_numeric_columns.max().max()
        df_min = df_numeric_columns.min().min()
        ranges = [
            ((df_max - df_min) * i) + df_min
            for i in bounds
        ]
        styles = []
        for i in range(1, len(bounds)):
            min_bound = ranges[i - 1]
            max_bound = ranges[i]
            backgroundColor = colorlover.scales[str(n_bins)]['seq'][colors][i - 1]
            color = 'white' if i > len(bounds) / 2. else 'inherit'

            for column in df_numeric_columns:
                styles.append({
                    'if': {
                        'filter_query': (
                                '{{{column}}} >= {min_bound}' +
                                (' && {{{column}}} < {max_bound}' if (i < len(bounds) - 1) else '')
                        ).format(column=column, min_bound=min_bound, max_bound=max_bound),
                        'column_id': column
                    },
                    'backgroundColor': backgroundColor,
                    'color': color
                })

        return styles

    def get_lineages_variants_table(self, data_source: str, lineages: List[str] = None, countries: List[str] = None):

        result = None
        if lineages is not None and len(lineages) == 1:
            query = self.queries.session.query(
                PrecomputedVariantsPerLineage.variant_id,
                PrecomputedVariantsPerLineage.count_observations,
                #Variant.variant_type,
                Variant.gene_name,
                Variant.hgvs_p,
                Variant.annotation_highest_impact
            )\
                .filter(and_(
                    PrecomputedVariantsPerLineage.source == data_source,
                    PrecomputedVariantsPerLineage.lineage == lineages[0]))

            if countries is not None and len(countries) > 0:
                query = query.filter(PrecomputedVariantsPerLineage.country.in_(countries))

            query = query.join(Variant, PrecomputedVariantsPerLineage.variant_id == Variant.variant_id)\
                .order_by(Variant.position.asc())
            data = pd.read_sql(query.statement, self.queries.session.bind)

            # adds together the counts from multiple countries
            data = data.groupby(["variant_id", "gene_name", "hgvs_p", "annotation_highest_impact"]).sum().reset_index()

            styles_counts = self.discrete_background_color_bins(data, columns=["count_observations"])

            fig = dash_table.DataTable(
                id='lineages-variants-table',
                data=data.to_dict('records'),
                columns=[
                    {"name": ["Gene"], "id": "gene_name"},
                    {"name": ["DNA mutation"], "id": "variant_id"},
                    {"name": ["Protein change"], "id": "hgvs_p"},
                    {"name": ["Annotation"], "id": "annotation_highest_impact"},
                    {"name": ["Count observations"], "id": "count_observations"}
                ],
                style_data_conditional=STYLES_STRIPPED + styles_counts,
                style_as_list_view=True,
                style_header=STYLE_HEADER,
                style_cell=STYLE_CELL,
                css=[{'selector': '.dash-cell div.dash-cell-value',
                      'rule': 'display: inline; white-space: inherit; overflow: inherit; text-overflow: inherit;'}],
                style_table={'overflowX': 'auto'},
                style_data={'whiteSpace': 'normal', 'height': 'auto'},
            )
            result = [fig, dcc.Markdown("**Mutations in lineage {lineage}**".format(lineage=lineages[0]))]

        return result



import os
from dataclasses import dataclass
from pathlib import Path
from logzero import logger
from covigator.configuration import Configuration
from covigator.pipeline.runner import run_command


@dataclass
class PipelineResult:
    # VCF
    lofreq_vcf: str
    ivar_vcf:  str
    gatk_vcf: str
    bcftools_vcf: str
    # QC
    fastp_qc: str
    vertical_coverage: str
    horizontal_coverage: str
    deduplication_metrics: str
    # pangolin results
    lofreq_pangolin: str
    ivar_pangolin: str
    gatk_pangolin: str
    bcftools_pangolin: str


class Pipeline:

    def __init__(self, config: Configuration):
        self.config = config
        assert self.config.reference_genome is not None and os.path.exists(self.config.reference_genome), \
            "Please configure the reference genome in the variable {}".format(self.config.ENV_COVIGATOR_REF_FASTA)

    def run(self, run_accession: str, fastq1: str, fastq2: str = None) -> PipelineResult:

        logger.info("Processing {} and {}".format(fastq1, fastq2))
        sample_data_folder = Path(fastq1).parent

        logger.info("Sample data folder: {}".format(sample_data_folder))

        lofreq_vcf = os.path.join(sample_data_folder, "{name}.lofreq.vcf.gz".format(name=run_accession))
        ivar_vcf = os.path.join(sample_data_folder, "{name}.ivar.vcf.gz".format(name=run_accession))
        gatk_vcf = os.path.join(sample_data_folder, "{name}.gatk.vcf.gz".format(name=run_accession))
        bcftools_vcf = os.path.join(sample_data_folder, "{name}.bcftools.vcf.gz".format(name=run_accession))

        lofreq_pangolin = os.path.join(sample_data_folder, "{name}.lofreq.pangolin.csv".format(name=run_accession))
        ivar_pangolin = os.path.join(sample_data_folder, "{name}.ivar.pangolin.csv".format(name=run_accession))
        gatk_pangolin = os.path.join(sample_data_folder, "{name}.gatk.pangolin.csv".format(name=run_accession))
        bcftools_pangolin = os.path.join(sample_data_folder, "{name}.bcftools.pangolin.csv".format(name=run_accession))

        output_qc = os.path.join(sample_data_folder, "{name}.fastp_stats.json".format(name=run_accession))
        output_vertical_coverage = os.path.join(sample_data_folder, "{name}.depth.tsv".format(name=run_accession))
        output_horizontal_coverage = os.path.join(sample_data_folder, "{name}.coverage.tsv".format(name=run_accession))
        deduplication_metrics = os.path.join(
            sample_data_folder, "{name}.deduplication_metrics.txt".format(name=run_accession))

        if not os.path.exists(lofreq_vcf) \
                or not os.path.exists(output_qc) \
                or not os.path.exists(output_horizontal_coverage) \
                or not os.path.exists(output_vertical_coverage) \
                or self.config.force_pipeline:

            command = "{nextflow} run {workflow} " \
                      "--fastq1 {fastq1} {fastq2} " \
                      "--output {output_folder} " \
                      "--name {name} " \
                      "--low_frequency_variant_threshold {af_low_frequency_thr} " \
                      "--subclonal_variant_threshold {af_subclonal_thr} " \
                      "--cpus {cpus} " \
                      "--memory {memory} " \
                      "--skip_bcftools " \
                      "--skip_gatk " \
                      "-profile conda " \
                      "-offline " \
                      "-work-dir {work_folder} " \
                      "-with-trace {trace_file}".format(
                nextflow=self.config.nextflow,
                fastq1=fastq1,
                fastq2="--fastq2 " + fastq2 if fastq2 else "",
                af_low_frequency_thr=self.config.low_coverage_threshold,
                af_subclonal_thr=self.config.subclonal_threshold,
                output_folder=sample_data_folder,
                name=run_accession,
                work_folder=self.config.temp_folder,
                workflow=self.config.workflow,
                trace_file=os.path.join(sample_data_folder, "nextflow_traces.txt"),
                cpus=self.config.workflow_cpus,
                memory=self.config.workflow_memory)
            run_command(command, sample_data_folder)

        return PipelineResult(
            # VCF
            lofreq_vcf=lofreq_vcf,
            ivar_vcf=ivar_vcf,
            gatk_vcf=gatk_vcf,
            bcftools_vcf=bcftools_vcf,
            # QC
            fastp_qc=output_qc,
            vertical_coverage=output_vertical_coverage,
            horizontal_coverage=output_horizontal_coverage,
            deduplication_metrics=deduplication_metrics,
            # pangolin
            lofreq_pangolin=lofreq_pangolin,
            ivar_pangolin=ivar_pangolin,
            gatk_pangolin=gatk_pangolin,
            bcftools_pangolin=bcftools_pangolin
        )



import dash_html_components as html
import dash_core_components as dcc
import dash_bootstrap_components as dbc

import covigator


def get_footer():
    tron_logo = "/assets/tron_logo_no_bg.png"
    return html.Footer(
        [
            dbc.Row([
                dbc.Col([
                    html.Br(),
                    html.P("CoVigator {} © 2021 TRON. All Rights Reserved".format(covigator.VERSION)),
                    html.P([
                        "GISAID data provided on this website are subject to GISAID’s ",
                        html.A("Terms and Conditions", href="https://www.gisaid.org/registration/terms-of-use/",
                               target="_blank")]),
                    html.A(html.Img(src=tron_logo, id="tron-logo"), href="https://tron-mainz.de",  target="_blank"),
                    html.Br(),
                    dcc.Markdown("""
                    TRON is an independent biopharmaceutical non-profit translational research organization pursuing 
                    new diagnostics and drugs for the treatment of cancer and other diseases with high medical need. 
                    We focus our transdisciplinary competencies in genomics and immunology to 
                    1) develop novel platforms for the identification and validation of “omics”-based biomarkers
                    and 2) for harnessing and modulating immune system components, for use in personalized therapies.
                    Partnering with academia and industry, TRON executes research at the leading edge to support 
                    innovative drug design for human health.
                    """),
                    html.P([
                        """
                    Intel is committed to accelerating access to technology that can combat the current pandemic 
                    and enable scientific discovery that better prepares our world for future crises. Funding for this 
                    solution was funded in part by """,
                        html.A(
                            "Intel’s Pandemic Response Technology Initiative",
                            href="https://newsroom.intel.com/news/intel-commits-technology-response-combat-coronavirus/",
                            target="_blank"),
                        """. For more 
                    information about healthcare solutions from Intel, visit intel.com/healthcare. For more 
                    information about Intel’s COVID-19 response, visit """,
                        html.A(
                            "intel.com/COVID-19",
                            href="https://www.intel.com/content/www/us/en/corporate-responsibility/covid-19-response.html",
                            target="_blank"),
                        "."
                        ]),
                    html.P([
                        html.A("DATA PROTECTION", href="https://tron-mainz.de/data-protection/",  target="_blank"), " | ",
                        html.A("IMPRINT", href="https://tron-mainz.de/imprint/",  target="_blank")])
                ], style={"color": "#003c78", 'margin-left': '15px'})]
                # this bit makes sure the footer sticks at the bottom
                #style={"position": "relative", "bottom": "0", "width": "100%", "overflow": "hidden", "height": "120px"}
            )
        ])



from typing import Union
from cyvcf2 import VCF, Variant
import os
import re
from sqlalchemy.exc import IntegrityError, InvalidRequestError
from sqlalchemy.orm import Session

from covigator import MISSENSE_VARIANT
from covigator.database.model import Variant as CovigatorVariant, VariantObservation, \
    SubclonalVariantObservation, SampleEna, SampleGisaid, DataSource, VariantType, GisaidVariantObservation, \
    LowFrequencyVariantObservation, GisaidVariant, SubclonalVariant, LowFrequencyVariant
from covigator.database.queries import Queries
from covigator.exceptions import CovigatorNotSupportedVariant


class VcfLoader:

    def load(self, vcf_file: str, run_accession: str, source: DataSource, session: Session):

        assert vcf_file is not None or vcf_file == "", "Missing VCF file provided to VcfLoader"
        assert os.path.exists(vcf_file) and os.path.isfile(vcf_file), "Non existing VCF file provided to VcfLoader"
        assert run_accession is not None or run_accession == "", "Missing sample"
        assert session is not None, "Missing DB session"

        observed_variants = []
        subclonal_observed_variants = []
        low_frequency_observed_variants = []
        specific_sample = Queries(session=session).find_sample_by_accession(
            run_accession=run_accession, source=source)
        assert specific_sample is not None, "Cannot find sample in database"

        # reads whole VCF in memory to count variants
        variants = [v for v in VCF(vcf_file)]

        # counts variants
        specific_sample.count_snvs = len([v for v in variants if v.FILTER is None and len(v.REF) == 1 and len(v.ALT[0]) == 1])
        specific_sample.count_deletions = len([v for v in variants if v.FILTER is None and len(v.REF) == 1 and len(v.ALT[0]) > 1])
        specific_sample.count_insertions = len([v for v in variants if v.FILTER is None and len(v.REF) > 1 and len(v.ALT[0]) == 1])

        if source == DataSource.ENA:
            specific_sample.count_subclonal_snvs = len([v for v in variants if v.FILTER == "SUBCLONAL"
                                        and len(v.REF) == 1 and len(v.ALT[0]) == 1])
            specific_sample.count_subclonal_deletions = len([v for v in variants if v.FILTER == "SUBCLONAL"
                                             and len(v.REF) > 1 and len(v.ALT[0]) == 1])
            specific_sample.count_subclonal_insertions = len([v for v in variants if v.FILTER == "SUBCLONAL"
                                              and len(v.REF) == 1 and len(v.ALT[0]) > 1])
            specific_sample.count_low_frequency_snvs = len([v for v in variants if v.FILTER  == "LOW_FREQUENCY"
                                        and len(v.REF) == 1 and len(v.ALT[0]) == 1])
            specific_sample.count_low_frequency_deletions = len([v for v in variants if v.FILTER == "LOW_FREQUENCY"
                                             and len(v.REF) > 1 and len(v.ALT[0]) == 1])
            specific_sample.count_low_frequency_insertions = len([v for v in variants if v.FILTER == "LOW_FREQUENCY"
                                              and len(v.REF) == 1 and len(v.ALT[0]) > 1])

        variant: Variant
        for variant in variants:
            if variant.FILTER is None or variant.FILTER in ["LOW_FREQUENCY", "SUBCLONAL"]:
                if source == DataSource.GISAID:
                    gisaid_variant = self._parse_variant(variant, GisaidVariant)
                    observed_variants.append(
                        self._parse_variant_observation(variant, specific_sample, source, gisaid_variant,
                                                        GisaidVariantObservation))
                    session.add(gisaid_variant)
                elif variant.FILTER is None:    # ENA clonal
                    # only stores clonal high quality variants in this table
                    ena_variant = self._parse_variant(variant, CovigatorVariant)
                    observed_variants.append(
                        self._parse_variant_observation(
                            variant, specific_sample, source, ena_variant, VariantObservation))
                    session.add(ena_variant)
                elif variant.FILTER == "SUBCLONAL":
                    subclonal_variant = self._parse_variant(variant, SubclonalVariant)
                    subclonal_observed_variants.append(
                        self._parse_variant_observation(
                            variant, specific_sample, source, subclonal_variant, SubclonalVariantObservation))
                    session.add(subclonal_variant)
                elif variant.FILTER == "LOW_FREQUENCY":
                    low_frequency_variant = self._parse_variant(variant, LowFrequencyVariant)
                    low_frequency_observed_variants.append(
                        self._parse_variant_observation(
                            variant, specific_sample, source, low_frequency_variant, LowFrequencyVariantObservation))
                    session.add(low_frequency_variant)
                try:
                    session.commit()
                except (IntegrityError, InvalidRequestError):
                    # do nothing the variant was just added by another process between merge and commit
                    session.rollback()
        session.add_all(observed_variants)
        session.add_all(subclonal_observed_variants)
        session.add_all(low_frequency_observed_variants)
        # NOTE: commit will happen afterwards when the job status is updated

    def _parse_variant(self, variant: Variant, klass) -> Union[CovigatorVariant, GisaidVariant]:
        parsed_variant = klass(
            chromosome=variant.CHROM,
            position=variant.POS,
            reference=variant.REF,
            # NOTE: because we only support haploid organisms we expect only one alternate,
            # TODO: support subclonal variants at some point
            alternate=variant.ALT[0],
            variant_type=self._get_variant_type(reference=variant.REF, alternate=variant.ALT[0]),
            length=self._get_variant_length(variant)
        )
        parsed_variant.variant_id = parsed_variant.get_variant_id()
        self._parse_snpeff_annotations(covigator_variant=parsed_variant, vcf_variant=variant)
        self._parse_additional_annotations(covigator_variant=parsed_variant, vcf_variant=variant)
        return parsed_variant

    def _parse_additional_annotations(
            self, covigator_variant: Union[CovigatorVariant, GisaidVariant, SubclonalVariant, LowFrequencyVariant],
            vcf_variant: Variant):
        covigator_variant.cons_hmm_sars_cov_2 = vcf_variant.INFO.get("CONS_HMM_SARS_COV_2")
        covigator_variant.cons_hmm_sarbecovirus = vcf_variant.INFO.get("CONS_HMM_SARBECOVIRUS")
        covigator_variant.cons_hmm_vertebrate_cov = vcf_variant.INFO.get("CONS_HMM_VERTEBRATE_COV")
        covigator_variant.pfam_name = vcf_variant.INFO.get("PFAM_NAME")
        covigator_variant.pfam_description = vcf_variant.INFO.get("PFAM_DESCRIPTION")

    def _parse_snpeff_annotations(
            self, covigator_variant: Union[CovigatorVariant, GisaidVariant, SubclonalVariant, LowFrequencyVariant],
            vcf_variant: Variant):
        ann = vcf_variant.INFO.get("ANN")
        if ann is not None:
            annotations = ann.split(",")
            annotation = annotations[0]  # NOTE: chooses arbitrarily the first annotation
            if "|" in annotation:
                values = annotation.split("|")
                covigator_variant.overlaps_multiple_genes = len(annotations) > 1
                covigator_variant.annotation = values[1].strip()
                covigator_variant.annotation_highest_impact = re.sub("&.*", "", covigator_variant.annotation)
                covigator_variant.annotation_impact = values[2].strip()
                covigator_variant.gene_name = values[3].strip()
                covigator_variant.gene_id = values[4].strip()
                covigator_variant.biotype = values[7].strip()
                covigator_variant.hgvs_c = values[9].strip()
                covigator_variant.hgvs_p = values[10].strip()
                covigator_variant.cdna_pos_length = values[11].strip()
                covigator_variant.cds_pos_length = values[12].strip()
                covigator_variant.aa_pos_length = values[13].strip()
                if covigator_variant.annotation == MISSENSE_VARIANT:
                    hgvs_pattern = re.compile(r"^p\.([a-zA-Z]{1,3})([0-9]+)([a-zA-Z]{1,3})$")
                    match = hgvs_pattern.match(covigator_variant.hgvs_p)
                    if match:
                        covigator_variant.reference_amino_acid = match.group(1)
                        covigator_variant.alternate_amino_acid = match.group(3)
                        covigator_variant.position_amino_acid = int(match.group(2))

    def _parse_variant_observation(
            self, variant: Variant, sample: Union[SampleEna, SampleGisaid], source: DataSource,
            covigator_variant: Union[CovigatorVariant, GisaidVariant], klass):

        dp4 = variant.INFO.get("DP4")
        return klass(
            sample=sample.run_accession,
            variant_id=covigator_variant.variant_id,
            chromosome=variant.CHROM,
            position=variant.POS,
            reference=variant.REF,
            # NOTE: because variants are normalized we only expect one alternate
            alternate=variant.ALT[0],
            quality=variant.QUAL,
            filter=variant.FILTER,
            dp=variant.INFO.get("vafator_dp"),
            dp4_ref_forward=dp4[0] if dp4 else None,
            dp4_ref_reverse=dp4[1] if dp4 else None,
            dp4_alt_forward=dp4[2] if dp4 else None,
            dp4_alt_reverse=dp4[3] if dp4 else None,
            vaf=variant.INFO.get("vafator_af"),
            ac=variant.INFO.get("vafator_ac"),
            strand_bias=variant.INFO.get("SB"),
            # denormalized fields
            annotation=covigator_variant.annotation,
            annotation_impact=covigator_variant.annotation_impact,
            biotype=covigator_variant.biotype,
            annotation_highest_impact=covigator_variant.annotation_highest_impact,
            gene_name=covigator_variant.gene_name,
            hgvs_p=covigator_variant.hgvs_p,
            hgvs_c=covigator_variant.hgvs_c,
            date=sample.first_created if source == DataSource.ENA else sample.collection_date,
            variant_type=covigator_variant.variant_type,
            length=self._get_variant_length(variant),
            reference_amino_acid=covigator_variant.reference_amino_acid,
            alternate_amino_acid=covigator_variant.alternate_amino_acid,
            position_amino_acid=covigator_variant.position_amino_acid,
            cons_hmm_sars_cov_2=covigator_variant.cons_hmm_sars_cov_2,
            cons_hmm_sarbecovirus=covigator_variant.cons_hmm_sarbecovirus,
            cons_hmm_vertebrate_cov=covigator_variant.cons_hmm_vertebrate_cov,
            pfam_name=covigator_variant.pfam_name,
            pfam_description=covigator_variant.pfam_description
        )

    def _get_variant_type(self, reference, alternate):
        if len(reference) == 1 and len(alternate) == 1:
            return VariantType.SNV
        elif len(reference) == len(alternate):
            return VariantType.MNV
        elif len(reference) == 1 and len(alternate) > 1:
            return VariantType.INSERTION
        elif len(reference) > 1 and len(alternate) == 1:
            return VariantType.DELETION
        else:
            raise CovigatorNotSupportedVariant

    def _get_variant_length(self, variant):
        return len(variant.ALT[0]) - len(variant.REF)



import os
import pathlib
import urllib.request as request
import hashlib
from logzero import logger
from covigator.configuration import Configuration
from covigator.database.model import SampleEna, SEPARATOR
import re

FTP_PROTOCOL = "ftp://"
PROTOCOL_REGEX = re.compile('[a-zA-Z0-9]+://')


class CovigatorMD5CheckSumError(Exception):
    pass


class Downloader:

    def __init__(self, config: Configuration):
        self.storage_folder = config.storage_folder
        pathlib.Path(self.storage_folder).mkdir(parents=True, exist_ok=True)
        assert os.path.exists(self.storage_folder), "Storage folder does not exist"

    def download(self, sample_ena: SampleEna) -> str:
        assert sample_ena.run_accession is not None, "A run accession is required as it is part of the file path"
        assert sample_ena.fastq_ftp is not None, "Cannot download empty URLs"
        assert sample_ena.fastq_md5 is not None, "Cannot do MD5 checks without MD5"
        fastqs = sample_ena.get_fastqs_ftp()
        md5s = sample_ena.get_fastqs_md5()
        paths = []
        for url, md5 in zip(fastqs, md5s):
            local_path = self._download_url(sample_ena, url, md5)
            paths.append(local_path)
        logger.info("Downloaded {}: {}".format(sample_ena.run_accession, sample_ena.fastq_ftp))
        return SEPARATOR.join(paths)

    def _download_url(self, sample: SampleEna, url, md5):
        """
        This method streams a file (probably large) from a URL and stores it in the local file system
        """
        if not re.match(PROTOCOL_REGEX, url):
            url = FTP_PROTOCOL + url
        local_filename = url.split('/')[-1]
        # NOTE: sample folder date/run_accession
        local_folder = sample.get_sample_folder(self.storage_folder)
        local_full_path = os.path.join(local_folder, local_filename)
        # avoids downloading the same files over and over
        if not os.path.exists(local_full_path):
            pathlib.Path(local_folder).mkdir(parents=True, exist_ok=True)
            request.urlretrieve(url, local_full_path)
            self._check_md5(local_full_path, md5)

        return local_full_path

    def _check_md5(self, filepath, md5):
        """
        This method reads a file (probably large) in chunks and computes the MD5 hash iteratively
        """
        file_hash = hashlib.md5()
        with open(filepath, "rb") as f:
            while True:
                chunk = f.read(8192)
                if not chunk:
                    break
                file_hash.update(chunk)
        if file_hash.hexdigest() != md5:
            raise CovigatorMD5CheckSumError("Failed MD5 check for file {} and MD5 {}".format(filepath, md5))
